#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import logging

from osc_lib.cli import parseractions
from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils
from osc_lib import utils as osc_utils

from manilaclient.common._i18n import _

LOG = logging.getLogger(__name__)


class CreateShareGroupSnapshot(command.ShowOne):
    """Create a share group snapshot."""
    _description = _(
        "Create a share group snapshot of the given share group")

    def get_parser(self, prog_name):
        parser = super(
            CreateShareGroupSnapshot, self).get_parser(prog_name)
        parser.add_argument(
            "share_group",
            metavar="<share-group>",
            help=_("Name or ID of the share group.")
        )
        parser.add_argument(
            "--name",
            metavar="<name>",
            default=None,
            help=_("Optional share group snapshot name. (Default=None)")
        )
        parser.add_argument(
            "--description",
            metavar="<description>",
            default=None,
            help=_("Optional share group snapshot description. "
                   "(Default=None)")
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            default=False,
            help=_('Wait for share group snapshot creation')
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        share_group = osc_utils.find_resource(
            share_client.share_groups,
            parsed_args.share_group)

        share_group_snapshot = share_client.share_group_snapshots.create(
            share_group,
            name=parsed_args.name,
            description=parsed_args.description,
        )
        if parsed_args.wait:
            if not osc_utils.wait_for_status(
                status_f=share_client.share_group_snapshots.get,
                res_id=share_group_snapshot.id,
                success_status=['available']
            ):
                LOG.error(_("ERROR: Share group snapshot is in error state."))

            share_group_snapshot = osc_utils.find_resource(
                share_client.share_group_snapshots,
                share_group_snapshot.id)

        data = share_group_snapshot._info
        data.pop('links', None)
        data.pop('members', None)

        return self.dict2columns(data)


class DeleteShareGroupSnapshot(command.Command):
    """Delete one or more share group snapshots."""
    _description = _("Delete one or more share group snapshot")

    def get_parser(self, prog_name):
        parser = super(DeleteShareGroupSnapshot, self).get_parser(prog_name)
        parser.add_argument(
            "share_group_snapshot",
            metavar="<share-group-snapshot>",
            nargs="+",
            help=_("Name or ID of the group snapshot(s) to delete")
        )
        parser.add_argument(
            "--force",
            action='store_true',
            default=False,
            help=_("Attempt to force delete the share group snapshot(s) "
                   "(Default=False) (Admin only).")
        )
        parser.add_argument(
            "--wait",
            action='store_true',
            default=False,
            help=_("Wait for share group snapshot deletion")
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        result = 0

        for share_group_snapshot in parsed_args.share_group_snapshot:
            try:
                share_group_snapshot_obj = osc_utils.find_resource(
                    share_client.share_group_snapshots,
                    share_group_snapshot)

                share_client.share_group_snapshots.delete(
                    share_group_snapshot_obj,
                    force=parsed_args.force)

                if parsed_args.wait:
                    if not osc_utils.wait_for_delete(
                            manager=share_client.share_group_snapshots,
                            res_id=share_group_snapshot_obj.id):
                        result += 1

            except Exception as e:
                result += 1
                LOG.error(
                    'Failed to delete a share group snapshot with '
                    f'name or ID {share_group_snapshot}: {e}')

        if result > 0:
            total = len(parsed_args.share_group_snapshot)
            msg = (f'{result} of {total} share group snapshots failed '
                   'to delete.')
            raise exceptions.CommandError(msg)


class ShowShareGroupSnapshot(command.ShowOne):
    """Display a share group snapshot"""
    _description = _(
        "Show details about a share group snapshot")

    def get_parser(self, prog_name):
        parser = super(ShowShareGroupSnapshot, self).get_parser(prog_name)
        parser.add_argument(
            "share_group_snapshot",
            metavar="<share-group-snapshot>",
            help=_("Name or ID of the share group snapshot to display")
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        share_group_snapshot = osc_utils.find_resource(
            share_client.share_group_snapshots,
            parsed_args.share_group_snapshot)

        data = share_group_snapshot._info
        data.pop('links', None)
        data.pop('members', None)

        return self.dict2columns(data)


class SetShareGroupSnapshot(command.Command):
    """Set share group snapshot properties."""
    _description = _("Set share group snapshot properties")

    def get_parser(self, prog_name):
        parser = super(SetShareGroupSnapshot, self).get_parser(prog_name)
        parser.add_argument(
            "share_group_snapshot",
            metavar="<share-group-snapshot>",
            help=_('Name or ID of the snapshot to set a property for')
        )
        parser.add_argument(
            "--name",
            metavar="<name>",
            default=None,
            help=_("Set a name to the snapshot.")
        )
        parser.add_argument(
            "--description",
            metavar="<description>",
            default=None,
            help=_("Set a description to the snapshot.")
        )
        parser.add_argument(
            "--status",
            metavar="<status>",
            choices=['available', 'error', 'creating',
                     'deleting', 'error_deleting'],
            help=_("Explicitly set the state of a share group snapshot"
                   "(Admin only). "
                   "Options include : available, error, creating, "
                   "deleting, error_deleting.")
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        result = 0

        share_group_snapshot = osc_utils.find_resource(
            share_client.share_group_snapshots,
            parsed_args.share_group_snapshot)

        kwargs = {}

        if parsed_args.name is not None:
            kwargs['name'] = parsed_args.name
        if parsed_args.description is not None:
            kwargs['description'] = parsed_args.description

        if kwargs:
            try:
                share_client.share_group_snapshots.update(
                    share_group_snapshot,
                    **kwargs
                )
            except Exception as e:
                result += 1
                LOG.error('Failed to set name or desciption for '
                          'share group snapshot with ID '
                          f'{share_group_snapshot.id}: {e}')

        if parsed_args.status:
            try:
                share_client.share_group_snapshots.reset_state(
                    share_group_snapshot,
                    parsed_args.status
                )
            except Exception as e:
                result += 1
                LOG.error('Failed to set status for '
                          'share group snapshot with ID '
                          f'{share_group_snapshot.id}: {e}')

        if result > 0:
            raise exceptions.CommandError(_(
                "One or more of the set operations failed"))


class UnsetShareGroupSnapshot(command.Command):
    """Unset a share group snapshot property."""
    _description = _("Unset a share group snapshot property")

    def get_parser(self, prog_name):
        parser = super(UnsetShareGroupSnapshot, self).get_parser(prog_name)
        parser.add_argument(
            "share_group_snapshot",
            metavar="<share-group-snapshot>",
            help=_("Name or ID of the group snapshot to unset a property of")
        )
        parser.add_argument(
            "--name",
            action='store_true',
            help=_("Unset share group snapshot name."),
        )
        parser.add_argument(
            "--description",
            action='store_true',
            help=_("Unset share group snapshot description."),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        share_group_snapshot = osc_utils.find_resource(
            share_client.share_group_snapshots,
            parsed_args.share_group_snapshot)

        kwargs = {}
        if parsed_args.name:
            # the SDK unsets name if it is an empty string
            kwargs['name'] = ''
        if parsed_args.description:
            # the SDK unsets description if it is an empty string
            kwargs['description'] = ''
        if kwargs:
            try:
                share_client.share_group_snapshots.update(
                    share_group_snapshot,
                    **kwargs
                )
            except Exception as e:
                raise exceptions.CommandError(
                    'Failed to unset name or description for '
                    f'share group snapshot : {e}')


class ListShareGroupSnapshot(command.Lister):
    """List share group snapshots."""
    _description = _("List share group snapshots")

    def get_parser(self, prog_name):
        parser = super(ListShareGroupSnapshot, self).get_parser(prog_name)
        parser.add_argument(
            "--all-projects",
            action='store_true',
            default=False,
            help=_("Display information from all projects (Admin only).")
        )
        parser.add_argument(
            "--name",
            metavar="<name>",
            default=None,
            help=_("Filter results by name.")
        )
        parser.add_argument(
            "--status",
            metavar="<status>",
            default=None,
            help=_("Filter results by status.")
        )
        parser.add_argument(
            "--share-group",
            metavar="<share-group>",
            default=None,
            help=_("Filter results by share group name or ID.")
        )
        parser.add_argument(
            "--limit",
            metavar="<limit>",
            type=int,
            default=None,
            action=parseractions.NonNegativeAction,
            help=_("Limit the number of share groups returned")
        )
        parser.add_argument(
            "--marker",
            metavar="<marker>",
            help=_("The last share group snapshot ID of the "
                   "previous page")
        )
        parser.add_argument(
            '--sort',
            metavar="<key>[:<direction>]",
            default='name:asc',
            help=_("Sort output by selected keys and directions(asc or desc) "
                   "(default: name:asc), multiple keys and directions can be "
                   "specified separated by comma")
        )
        parser.add_argument(
            "--detailed",
            action="store_true",
            help=_("Show detailed information about share group snapshot. ")
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        share_group_id = None
        if parsed_args.share_group:
            share_group_id = osc_utils.find_resource(
                share_client.share_groups,
                parsed_args.share_group).id

        columns = [
            'ID',
            'Name',
            'Status',
            'Description',
        ]

        search_opts = {
            'all_tenants': parsed_args.all_projects,
            'name': parsed_args.name,
            'status': parsed_args.status,
            'share_group_id': share_group_id,
            'limit': parsed_args.limit,
            'offset': parsed_args.marker,
        }

        if parsed_args.detailed:
            columns.extend([
                'Created At',
                'Share Group ID',
            ])

        if parsed_args.all_projects:
            columns.append('Project ID')

        share_group_snapshots = share_client.share_group_snapshots.list(
            search_opts=search_opts)

        share_group_snapshots = utils.sort_items(
            share_group_snapshots, parsed_args.sort, str)

        data = (
            osc_utils.get_dict_properties(share_group_snapshot._info, columns)
            for share_group_snapshot in share_group_snapshots
        )

        return (columns, data)


class ListShareGroupSnapshotMembers(command.Lister):
    """List members for share group snapshot."""
    _description = _("List members of share group snapshot")

    def get_parser(self, prog_name):
        parser = super(
            ListShareGroupSnapshotMembers, self).get_parser(prog_name)
        parser.add_argument(
            "share_group_snapshot",
            metavar="<share-group-snapshot>",
            help=_("Name or ID of the group snapshot to list members for")
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        columns = ['Share ID', 'Size']

        share_group_snapshot = osc_utils.find_resource(
            share_client.share_group_snapshots,
            parsed_args.share_group_snapshot)

        data = (
            osc_utils.get_dict_properties(member, columns)
            for member in share_group_snapshot._info.get('members', [])
        )

        return (columns, data)



# Copyright 2015 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
import testtools

from tempest.lib.common.utils import data_utils
from tempest.lib import exceptions

from manilaclient.common import constants
from manilaclient import config
from manilaclient.tests.functional import base
from manilaclient.tests.functional import utils

CONF = config.CONF


@ddt.ddt
class SharesReadWriteBase(base.BaseTestCase):
    protocol = None

    def setUp(self):
        super(SharesReadWriteBase, self).setUp()
        if self.protocol not in CONF.enable_protocols:
            message = "%s tests are disabled" % self.protocol
            raise self.skipException(message)
        self.name = data_utils.rand_name('autotest_share_name')
        self.description = data_utils.rand_name('autotest_share_description')

        # NOTE(vponomaryov): following share is used only in one test
        # until tests for snapshots appear.
        self.share = self.create_share(
            share_protocol=self.protocol,
            size=1,
            name=self.name,
            description=self.description,
            client=self.get_user_client())

    def test_create_delete_share(self):
        name = data_utils.rand_name('autotest_share_name')

        create = self.create_share(
            self.protocol, name=name, client=self.user_client)

        self.assertEqual("creating", create['status'])
        self.assertEqual(name, create['name'])
        self.assertEqual('1', create['size'])
        self.assertEqual(self.protocol.upper(), create['share_proto'])

        self.user_client.delete_share(create['id'])

        self.user_client.wait_for_share_deletion(create['id'])

    def test_create_update_share(self):
        name = data_utils.rand_name('autotest_share_name')
        new_name = 'new_' + name
        description = data_utils.rand_name('autotest_share_description')
        new_description = 'new_' + description

        create = self.create_share(
            self.protocol, name=name, description=description,
            client=self.user_client)

        self.assertEqual(name, create['name'])
        self.assertEqual(description, create['description'])
        self.assertEqual('False', create['is_public'])

        self.user_client.update_share(
            create['id'], name=new_name, description=new_description)
        get = self.user_client.get_share(create['id'])

        self.assertEqual(new_name, get['name'])
        self.assertEqual(new_description, get['description'])
        self.assertEqual('False', get['is_public'])

        # only admins operating at system scope can set a share to be public
        self.admin_client.update_share(create['id'], is_public=True)
        get = self.user_client.get_share(create['id'])

        self.assertEqual(new_name, get['name'])
        self.assertEqual(new_description, get['description'])
        self.assertEqual('True', get['is_public'])

    def test_get_share(self):
        get = self.user_client.get_share(self.share['id'])

        self.assertEqual(self.name, get['name'])
        self.assertEqual(self.description, get['description'])
        self.assertEqual('1', get['size'])
        self.assertEqual(self.protocol.upper(), get['share_proto'])

    def test_create_delete_with_wait(self):
        name = data_utils.rand_name('share-with-wait-%s')
        description = data_utils.rand_name('we-wait-until-share-is-ready')

        share_1, share_2 = (
            self.create_share(self.protocol, name=(name % num),
                              description=description,
                              use_wait_option=True,
                              client=self.user_client)
            for num in range(0, 2)
        )

        self.assertEqual("available", share_1['status'])
        self.assertEqual("available", share_2['status'])

        self.delete_share([share_1['id'], share_2['id']],
                          wait=True,
                          client=self.user_client)

        for share in (share_1, share_2):
            self.assertRaises(exceptions.NotFound,
                              self.user_client.get_share,
                              share['id'])

    def test_create_soft_delete_and_restore_share(self):
        self.skip_if_microversion_not_supported('2.69')
        microversion = '2.69'
        description = data_utils.rand_name('we-wait-until-share-is-ready')

        share = self.create_share(self.protocol,
                                  name='share_name',
                                  description=description,
                                  use_wait_option=True,
                                  client=self.user_client)

        self.assertEqual("available", share['status'])

        # soft delete the share to recycle bin
        self.soft_delete_share([share['id']], client=self.user_client,
                               microversion=microversion)
        self.user_client.wait_for_share_soft_deletion(share['id'])

        # get shares list in recycle bin
        result = self.user_client.list_shares(is_soft_deleted=True,
                                              microversion=microversion)
        share_ids = [sh['ID'] for sh in result]
        # check share is in recycle bin
        self.assertIn(share['id'], share_ids)

        # restore the share from recycle bin
        self.restore_share([share['id']], client=self.user_client,
                           microversion=microversion)
        self.user_client.wait_for_share_restore(share['id'])

        result1 = self.user_client.list_shares(is_soft_deleted=True,
                                               microversion=microversion)
        share_ids1 = [sh['ID'] for sh in result1]
        # check share not in recycle bin
        self.assertNotIn(share['id'], share_ids1)


@ddt.ddt
class SharesTestMigration(base.BaseTestCase):

    def setUp(self):
        super(SharesTestMigration, self).setUp()

        self.old_type = self.create_share_type(
            data_utils.rand_name('test_share_type'),
            driver_handles_share_servers=True)
        self.new_type = self.create_share_type(
            data_utils.rand_name('test_share_type'),
            driver_handles_share_servers=True)
        self.error_type = self.create_share_type(
            data_utils.rand_name('test_share_type'),
            driver_handles_share_servers=True,
            extra_specs={'cause_error': 'no_valid_host'})

        self.old_share_net = self.get_user_client().get_share_network(
            self.get_user_client().share_network)
        share_net_info = (
            utils.get_default_subnet(self.get_user_client(),
                                     self.old_share_net['id'])
            if utils.share_network_subnets_are_supported()
            else self.old_share_net)
        self.new_share_net = self.create_share_network(
            neutron_net_id=share_net_info['neutron_net_id'],
            neutron_subnet_id=share_net_info['neutron_subnet_id'])

    @utils.skip_if_microversion_not_supported('2.22')
    @ddt.data('migration_error', 'migration_success', 'None')
    def test_reset_task_state(self, state):

        share = self.create_share(
            share_protocol='nfs',
            size=1,
            name=data_utils.rand_name('autotest_share_name'),
            client=self.get_user_client(),
            share_type=self.old_type['ID'],
            share_network=self.old_share_net['id'],
            wait_for_creation=True)
        share = self.user_client.get_share(share['id'])

        self.admin_client.reset_task_state(share['id'], state)
        share = self.user_client.get_share(share['id'])
        self.assertEqual(state, share['task_state'])

    @utils.skip_if_microversion_not_supported('2.29')
    @testtools.skipUnless(
        CONF.run_migration_tests, 'Share migration tests are disabled.')
    @ddt.data('cancel', 'success', 'error')
    def test_full_migration(self, test_type):
        # We are testing with DHSS=True only because it allows us to specify
        # new_share_network.

        share = self.create_share(
            share_protocol='nfs',
            size=1,
            name=data_utils.rand_name('autotest_share_name'),
            client=self.get_user_client(),
            share_type=self.old_type['ID'],
            share_network=self.old_share_net['id'],
            wait_for_creation=True)
        share = self.admin_client.get_share(share['id'])

        pools = self.admin_client.pool_list(detail=True)

        dest_pool = utils.choose_matching_backend(
            share, pools, self.new_type)

        self.assertIsNotNone(dest_pool)

        source_pool = share['host']

        new_type = self.new_type
        if test_type == 'error':
            statuses = constants.TASK_STATE_MIGRATION_ERROR
            new_type = self.error_type
        else:
            statuses = (constants.TASK_STATE_MIGRATION_DRIVER_PHASE1_DONE,
                        constants.TASK_STATE_DATA_COPYING_COMPLETED)

        self.admin_client.migration_start(
            share['id'], dest_pool, writable=True, nondisruptive=False,
            preserve_metadata=True, preserve_snapshots=True,
            force_host_assisted_migration=False,
            new_share_network=self.new_share_net['id'],
            new_share_type=new_type['ID'])

        share = self.admin_client.wait_for_migration_task_state(
            share['id'], dest_pool, statuses)

        progress = self.admin_client.migration_get_progress(share['id'])
        self.assertEqual('100', progress['total_progress'])

        self.assertEqual(source_pool, share['host'])
        self.assertEqual(self.old_type['ID'], share['share_type'])
        self.assertEqual(self.old_share_net['id'], share['share_network_id'])

        if test_type == 'error':
            self.assertEqual(statuses, progress['task_state'])
        else:
            if test_type == 'success':
                self.admin_client.migration_complete(share['id'])
                statuses = constants.TASK_STATE_MIGRATION_SUCCESS
            elif test_type == 'cancel':
                self.admin_client.migration_cancel(share['id'])
                statuses = constants.TASK_STATE_MIGRATION_CANCELLED

            share = self.admin_client.wait_for_migration_task_state(
                share['id'], dest_pool, statuses)
            progress = self.admin_client.migration_get_progress(share['id'])
            self.assertEqual(statuses, progress['task_state'])
            if test_type == 'success':
                self.assertEqual(dest_pool, share['host'])
                self.assertEqual(new_type['ID'], share['share_type'])
                self.assertEqual(self.new_share_net['id'],
                                 share['share_network_id'])
            else:
                self.assertEqual(source_pool, share['host'])
                self.assertEqual(self.old_type['ID'], share['share_type'])
                self.assertEqual(self.old_share_net['id'],
                                 share['share_network_id'])


class NFSSharesReadWriteTest(SharesReadWriteBase):
    protocol = 'nfs'


class CIFSSharesReadWriteTest(SharesReadWriteBase):
    protocol = 'cifs'


class GlusterFSSharesReadWriteTest(SharesReadWriteBase):
    protocol = 'glusterfs'


class HDFSSharesReadWriteTest(SharesReadWriteBase):
    protocol = 'hdfs'


class MAPRFSSharesReadWriteTest(SharesReadWriteBase):
    protocol = 'maprfs'



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
A fake server that "responds" to API methods with pre-canned responses.

All of these responses come from the spec, so if for some reason the spec's
wrong the tests might raise AssertionError. I've indicated in comments the
places where actual behavior differs from the spec.
"""


def assert_has_keys(dictonary, required=None, optional=None):
    if required is None:
        required = []
    if optional is None:
        optional = []
    for k in required:
        try:
            assert k in dictonary
        except AssertionError:
            extra_keys = set(dictonary).difference(set(required + optional))
            raise AssertionError("found unexpected keys: %s" %
                                 list(extra_keys))


class FakeClient(object):

    def assert_called(self, method, url, body=None, pos=-1, **kwargs):
        """Assert than an API method was just called."""
        expected = (method, url)
        called = self.client.callstack[pos][0:2]

        assert self.client.callstack, ("Expected %s %s but no calls "
                                       "were made." % expected)

        assert expected == called, 'Expected %s %s; got %s %s' % (
            expected + called)

        if body is not None:
            actual = self.client.callstack[pos][2]
            assert actual == body, "Expected %s; got %s" % (body, actual)

    def assert_called_anytime(self, method, url, body=None,
                              clear_callstack=True):
        """Assert than an API method was called anytime in the test."""
        expected = (method, url)

        assert self.client.callstack, ("Expected %s %s but no calls "
                                       "were made." % expected)

        found = False
        for entry in self.client.callstack:
            if expected == entry[0:2]:
                found = True
                break

        assert found, 'Expected %s %s; got %s' % (
            expected, self.client.callstack)

        if body is not None:
            try:
                assert entry[2] == body
            except AssertionError:
                print(entry[2])
                print("!=")
                print(body)
                raise
        if clear_callstack:
            self.client.callstack = []

    def clear_callstack(self):
        self.client.callstack = []

    def authenticate(self):
        pass



# Copyright 2016 Mirantis, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import ddt

from manilaclient import api_versions
from manilaclient.tests.unit import utils
from manilaclient.v2 import availability_zones


@ddt.ddt
class AvailabilityZoneTest(utils.TestCase):

    def _get_manager(self, microversion):
        version = api_versions.APIVersion(microversion)
        mock_microversion = mock.Mock(api_version=version)
        return availability_zones.AvailabilityZoneManager(
            api=mock_microversion)

    def _get_resource_path(self, microversion):
        if (api_versions.APIVersion(microversion) >
                api_versions.APIVersion("2.6")):
            return availability_zones.RESOURCE_PATH
        return availability_zones.RESOURCE_PATH_LEGACY

    @ddt.data("2.6", "2.7", api_versions.MIN_VERSION, api_versions.MAX_VERSION)
    def test_list(self, microversion):
        manager = self._get_manager(microversion)
        resource_path = self._get_resource_path(microversion)
        self.mock_object(manager, "_list")

        result = manager.list()

        manager._list.assert_called_once_with(
            resource_path, availability_zones.RESOURCE_NAME)
        self.assertEqual(manager._list.return_value, result)

    def test_representation(self):
        resource = availability_zones.AvailabilityZone(None, {})
        resource.id = "9d21a755-b5ca-453e-a155-ee9c9066d909"
        expected = "<AvailabilityZone: 9d21a755-b5ca-453e-a155-ee9c9066d909>"
        self.assertEqual(repr(resource), expected)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from manilaclient import config
from manilaclient.tests.functional import base
from manilaclient.tests.functional import utils

CONF = config.CONF


@utils.skip_if_microversion_not_supported('2.72')
class ShareReplicasTest(base.BaseTestCase):

    def _create_share_and_replica(self):
        replication_type = CONF.replication_type
        share_type = self.create_share_type(
            driver_handles_share_servers=True,
            extra_specs={'replication_type': replication_type})
        share_network = self.create_share_network()
        share = self.create_share(
            share_type=share_type['ID'],
            share_network=share_network['id'],
            client=self.get_user_client())
        share_replica = self.create_share_replica(
            share['id'],
            share_network=share_network['id'],
            wait_for_creation=True,
            client=self.get_user_client())
        return share, share_replica

    def test_share_replica_create(self):
        share, share_replica = self._create_share_and_replica()
        self.assertEqual(share['id'], share_replica['share_id'])

    def test_share_replica_delete(self):
        share, share_replica = self._create_share_and_replica()
        self.user_client.delete_share_replica(share_replica['id'])
        self.user_client.wait_for_share_replica_deletion(share_replica['id'])



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#
from osc_lib import exceptions
from osc_lib import utils as oscutils

from manilaclient import api_versions
from manilaclient.osc.v2 import security_services as osc_security_services
from manilaclient.tests.unit.osc import osc_utils
from manilaclient.tests.unit.osc.v2 import fakes as manila_fakes


class TestShareSecurityService(manila_fakes.TestShare):

    def setUp(self):
        super(TestShareSecurityService, self).setUp()

        self.security_services_mock = (
            self.app.client_manager.share.security_services)
        self.security_services_mock.reset_mock()

        self.share_networks_mock = self.app.client_manager.share.share_networks
        self.share_networks_mock.reset_mock()

        self.app.client_manager.share.api_version = api_versions.APIVersion(
            api_versions.MAX_VERSION
        )


class TestShareSecurityServiceCreate(TestShareSecurityService):

    def setUp(self):
        super(TestShareSecurityServiceCreate, self).setUp()

        self.security_service = manila_fakes.FakeShareSecurityService \
            .create_fake_security_service()
        self.security_services_mock.create.return_value = self.security_service
        self.cmd = osc_security_services.CreateShareSecurityService(
            self.app, None)

        self.data = self.security_service._info.values()
        self.columns = self.security_service._info.keys()

    def test_share_security_service_create_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_utils.ParserException,
                          self.check_parser, self.cmd, arglist, verifylist)

    def test_share_security_service_create(self):
        arglist = [
            self.security_service.type,
            '--dns-ip', self.security_service.dns_ip,
            '--ou', self.security_service.ou,
            '--server', self.security_service.server,
            '--domain', self.security_service.domain,
            '--user', self.security_service.user,
            '--password', self.security_service.password,
            '--name', self.security_service.name,
            '--description', self.security_service.description
        ]
        verifylist = [
            ('type', self.security_service.type),
            ('dns_ip', self.security_service.dns_ip),
            ('ou', self.security_service.ou),
            ('server', self.security_service.server),
            ('domain', self.security_service.domain),
            ('user', self.security_service.user),
            ('password', self.security_service.password),
            ('name', self.security_service.name),
            ('description', self.security_service.description)
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.security_services_mock.create.assert_called_with(
            self.security_service.type,
            dns_ip=self.security_service.dns_ip,
            server=self.security_service.server,
            domain=self.security_service.domain,
            user=self.security_service.user,
            password=self.security_service.password,
            name=self.security_service.name,
            description=self.security_service.description,
            ou=self.security_service.ou
        )

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_share_security_service_create_api_version_exception(self):
        self.app.client_manager.share.api_version = api_versions.APIVersion(
            '2.43'
        )

        arglist = [
            self.security_service.type,
            '--ou', self.security_service.ou,
        ]
        verifylist = [
            ('type', self.security_service.type),
            ('ou', self.security_service.ou),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args)


class TestShareSecurityServiceDelete(TestShareSecurityService):

    def setUp(self):
        super(TestShareSecurityServiceDelete, self).setUp()

        self.security_service = manila_fakes.FakeShareSecurityService \
            .create_fake_security_service()
        self.security_services_mock.get.return_value = self.security_service

        self.security_services = manila_fakes.FakeShareSecurityService \
            .create_fake_security_services()

        self.cmd = osc_security_services.DeleteShareSecurityService(
            self.app, None)

    def test_share_security_service_delete_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_utils.ParserException,
                          self.check_parser, self.cmd, arglist, verifylist)

    def test_share_security_service_delete(self):
        arglist = [
            self.security_services[0].id,
            self.security_services[1].id,
        ]
        verifylist = [
            ('security_service', [self.security_services[0].id,
                                  self.security_services[1].id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.assertEqual(self.security_services_mock.delete.call_count,
                         len(self.security_services))
        self.assertIsNone(result)

    def test_share_security_service_delete_exception(self):
        arglist = [
            self.security_services[0].id,
        ]
        verifylist = [
            ('security_service', [self.security_services[0].id]),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.security_services_mock.delete.side_effect = \
            exceptions.CommandError()
        self.assertRaises(exceptions.CommandError,
                          self.cmd.take_action,
                          parsed_args)


class TestShareSecurityServiceShow(TestShareSecurityService):

    def setUp(self):
        super(TestShareSecurityServiceShow, self).setUp()

        self.security_service = manila_fakes.FakeShareSecurityService \
            .create_fake_security_service()
        self.security_services_mock.get.return_value = self.security_service

        self.cmd = osc_security_services.ShowShareSecurityService(
            self.app, None)

        self.data = self.security_service._info.values()
        self.columns = self.security_service._info.keys()

    def test_share_security_service_show_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_utils.ParserException,
                          self.check_parser, self.cmd, arglist, verifylist)

    def test_share_security_service_show(self):
        arglist = [
            self.security_service.id
        ]
        verifylist = [
            ('security_service', self.security_service.id)
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.security_services_mock.get.assert_called_with(
            self.security_service.id
        )

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)


class TestShareSecurityServiceSet(TestShareSecurityService):

    def setUp(self):
        super(TestShareSecurityServiceSet, self).setUp()

        self.security_service = manila_fakes.FakeShareSecurityService \
            .create_fake_security_service(methods={'update': None})
        self.security_services_mock.get.return_value = self.security_service
        self.cmd = osc_security_services.SetShareSecurityService(
            self.app, None)

    def test_share_security_service_set_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_utils.ParserException,
                          self.check_parser, self.cmd, arglist, verifylist)

    def test_share_security_service_set(self):
        arglist = [
            self.security_service.id,
            '--dns-ip', self.security_service.dns_ip,
            '--ou', self.security_service.ou,
            '--server', self.security_service.server,
            '--domain', self.security_service.domain,
            '--user', self.security_service.user,
            '--password', self.security_service.password,
            '--name', self.security_service.name,
            '--description', self.security_service.description
        ]
        verifylist = [
            ('security_service', self.security_service.id),
            ('dns_ip', self.security_service.dns_ip),
            ('ou', self.security_service.ou),
            ('server', self.security_service.server),
            ('domain', self.security_service.domain),
            ('user', self.security_service.user),
            ('password', self.security_service.password),
            ('name', self.security_service.name),
            ('description', self.security_service.description)
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.security_service.update.assert_called_with(
            dns_ip=self.security_service.dns_ip,
            server=self.security_service.server,
            domain=self.security_service.domain,
            user=self.security_service.user,
            password=self.security_service.password,
            name=self.security_service.name,
            description=self.security_service.description,
            ou=self.security_service.ou
        )
        self.assertIsNone(result)

    def test_share_security_service_set_exception(self):
        arglist = [
            self.security_service.id,
            '--name', self.security_service.name,
        ]
        verifylist = [
            ('security_service', self.security_service.id),
            ('name', self.security_service.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.security_service.update.side_effect = \
            exceptions.CommandError()
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args)

    def test_share_security_service_set_api_version_exception(self):
        self.app.client_manager.share.api_version = api_versions.APIVersion(
            '2.43'
        )

        arglist = [
            self.security_service.id,
            '--ou', self.security_service.ou,
        ]
        verifylist = [
            ('security_service', self.security_service.id),
            ('ou', self.security_service.ou),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args)


class TestShareSecurityServiceUnset(TestShareSecurityService):

    def setUp(self):
        super(TestShareSecurityServiceUnset, self).setUp()

        self.security_service = manila_fakes.FakeShareSecurityService \
            .create_fake_security_service(methods={'update': None})
        self.security_services_mock.get.return_value = self.security_service
        self.cmd = osc_security_services.UnsetShareSecurityService(
            self.app, None)

    def test_share_security_service_unset_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_utils.ParserException,
                          self.check_parser, self.cmd, arglist, verifylist)

    def test_share_security_service_unset(self):
        arglist = [
            self.security_service.id,
            '--dns-ip',
            '--ou',
            '--server',
            '--domain',
            '--user',
            '--password',
            '--name',
            '--description',
        ]
        verifylist = [
            ('security_service', self.security_service.id),
            ('dns_ip', True),
            ('ou', True),
            ('server', True),
            ('domain', True),
            ('user', True),
            ('password', True),
            ('name', True),
            ('description', True)
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.security_service.update.assert_called_with(
            dns_ip='',
            server='',
            domain='',
            user='',
            password='',
            name='',
            description='',
            ou=''
        )
        self.assertIsNone(result)

    def test_share_security_service_unset_exception(self):
        arglist = [
            self.security_service.id,
            '--name',
        ]
        verifylist = [
            ('security_service', self.security_service.id),
            ('name', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.security_service.update.side_effect = \
            exceptions.CommandError()
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args)

    def test_share_security_service_unset_api_version_exception(self):
        self.app.client_manager.share.api_version = api_versions.APIVersion(
            '2.43'
        )

        arglist = [
            self.security_service.id,
            '--ou',
        ]
        verifylist = [
            ('security_service', self.security_service.id),
            ('ou', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args)


class TestShareSecurityServiceList(TestShareSecurityService):

    columns = [
        'ID',
        'Name',
        'Status',
        'Type',
    ]

    def setUp(self):
        super(TestShareSecurityServiceList, self).setUp()

        self.share_network = (
            manila_fakes.FakeShareNetwork.create_one_share_network())
        self.share_networks_mock.get.return_value = self.share_network
        self.services_list = manila_fakes.FakeShareSecurityService \
            .create_fake_security_services()
        self.security_services_mock.list.return_value = self.services_list
        self.values = (oscutils.get_dict_properties(
            i._info, self.columns) for i in self.services_list)

        self.cmd = osc_security_services.ListShareSecurityService(
            self.app, None)

    def test_share_security_service_list_no_args(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.security_services_mock.list.assert_called_with(
            search_opts={
                'all_tenants': False,
                'status': None,
                'name': None,
                'type': None,
                'user': None,
                'dns_ip': None,
                'server': None,
                'domain': None,
                'offset': None,
                'limit': None,
            },
            detailed=False)
        self.assertEqual(self.columns, columns)
        self.assertEqual(list(self.values), list(data))

    def test_share_security_service_list(self):
        arglist = [
            '--share-network', self.share_network.id,
            '--status', self.services_list[0].status,
            '--name', self.services_list[0].name,
            '--type', self.services_list[0].type,
            '--user', self.services_list[0].user,
            '--dns-ip', self.services_list[0].dns_ip,
            '--ou', self.services_list[0].ou,
            '--server', self.services_list[0].server,
            '--domain', self.services_list[0].domain,
            '--limit', '1',
        ]
        verifylist = [
            ('share_network', self.share_network.id),
            ('status', self.services_list[0].status),
            ('name', self.services_list[0].name),
            ('type', self.services_list[0].type),
            ('user', self.services_list[0].user),
            ('dns_ip', self.services_list[0].dns_ip),
            ('ou', self.services_list[0].ou),
            ('server', self.services_list[0].server),
            ('domain', self.services_list[0].domain),
            ('limit', 1),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.security_services_mock.list.assert_called_with(
            search_opts={
                'all_tenants': False,
                'status': self.services_list[0].status,
                'name': self.services_list[0].name,
                'type': self.services_list[0].type,
                'user': self.services_list[0].user,
                'dns_ip': self.services_list[0].dns_ip,
                'server': self.services_list[0].server,
                'domain': self.services_list[0].domain,
                'offset': None,
                'limit': 1,
                'ou': self.services_list[0].ou,
                'share_network_id': self.share_network.id,
            },
            detailed=False)

        self.assertEqual(self.columns, columns)
        self.assertEqual(list(self.values), list(data))

    def test_share_security_service_list_ou_api_version_exception(self):
        self.app.client_manager.share.api_version = api_versions.APIVersion(
            '2.43'
        )
        arglist = [
            '--ou', self.services_list[0].ou,
        ]
        verifylist = [
            ('ou', self.services_list[0].ou),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args)

    def test_share_security_service_list_detail_all_projects(self):
        arglist = [
            '--all-projects',
            '--detail'
        ]
        verifylist = [
            ('all_projects', True),
            ('detail', True),
        ]
        columns_detail = self.columns.copy()
        columns_detail.append('Project ID')
        columns_detail.append('Share Networks')

        values_detail = (oscutils.get_dict_properties(
            i._info, columns_detail) for i in self.services_list)

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.security_services_mock.list.assert_called_with(
            search_opts={
                'all_tenants': True,
                'status': None,
                'name': None,
                'type': None,
                'user': None,
                'dns_ip': None,
                'server': None,
                'domain': None,
                'offset': None,
                'limit': None,
            },
            detailed=True)
        self.assertEqual(columns_detail, columns)
        self.assertEqual(list(values_detail), list(data))



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils

from manilaclient import api_versions
from manilaclient.common._i18n import _


class QuotaSet(command.Command):
    """Set quotas for a project or project/user or project/share-type.

    It can be used to set the default class for all projects.
    """

    _description = _("Set Quota for a project, or project/user or "
                     "project/share-type or a class.")

    def get_parser(self, prog_name):
        parser = super(QuotaSet, self).get_parser(prog_name)
        quota_type = parser.add_mutually_exclusive_group()
        parser.add_argument(
            'project',
            metavar='<project/class>',
            help=_("A project (name/ID) or a class (e.g.: default).")
        )
        quota_type.add_argument(
            '--class',
            dest='quota_class',
            action='store_true',
            default=False,
            help=_("Update class quota to all projects. "
                   "Mutually exclusive with '--user' and '--share-type'.")
        )
        quota_type.add_argument(
            '--user',
            metavar='<user>',
            default=None,
            help=_("Name or ID of a user to set the quotas for. "
                   "Mutually exclusive with '--share-type' and '--class'.")
        )
        quota_type.add_argument(
            '--share-type',
            metavar='<share-type>',
            type=str,
            default=None,
            help=_("Name or ID of a share type to set the quotas for. "
                   "Mutually exclusive with '--user' and '--class'. "
                   "Available only for microversion >= 2.39")
        )
        parser.add_argument(
            '--shares',
            metavar='<shares>',
            type=int,
            default=None,
            help=_('New value for the "shares" quota.')
        )
        parser.add_argument(
            '--snapshots',
            metavar='<snapshots>',
            type=int,
            default=None,
            help=_('New value for the "snapshots" quota.')
        )
        parser.add_argument(
            '--gigabytes',
            metavar='<gigabytes>',
            type=int,
            default=None,
            help=_('New value for the "gigabytes" quota.')
        )
        parser.add_argument(
            '--snapshot-gigabytes',
            metavar='<snapshot-gigabytes>',
            type=int,
            default=None,
            help=_('New value for the "snapshot-gigabytes" quota.')
        )
        parser.add_argument(
            '--share-networks',
            metavar='<share-networks>',
            type=int,
            default=None,
            help=_('New value for the "share-networks" quota.')
        )
        parser.add_argument(
            '--share-groups',
            metavar='<share-groups>',
            type=int,
            default=None,
            help=_('New value for the "share-groups" quota. '
                   'Available only for microversion >= 2.40')
        )
        parser.add_argument(
            '--share-group-snapshots',
            metavar='<share-group-snapshots>',
            type=int,
            default=None,
            help=_('New value for the "share-group-snapshots" quota. '
                   'Available only for microversion >= 2.40')
        )
        parser.add_argument(
            '--share-replicas',
            metavar='<share-replicas>',
            type=int,
            default=None,
            help=_("Number of share replicas. "
                   "Available only for microversion >= 2.53")
        )
        parser.add_argument(
            '--replica-gigabytes',
            metavar='<replica-gigabytes>',
            type=int,
            default=None,
            help=_("Capacity of share replicas in total. "
                   "Available only for microversion >= 2.53")
        )
        parser.add_argument(
            '--per-share-gigabytes',
            metavar='<per-share-gigabytes>',
            type=int,
            default=None,
            help=_("New value for the 'per-share-gigabytes' quota."
                   "Available only for microversion >= 2.62")
        )
        parser.add_argument(
            '--force',
            dest='force',
            action="store_true",
            default=None,
            help=_('Force update the quota. '
                   'Not applicable for class update.')
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        identity_client = self.app.client_manager.identity

        user_id = None
        if parsed_args.user:
            user_id = utils.find_resource(
                identity_client.users,
                parsed_args.user).id

        kwargs = {
            "shares": parsed_args.shares,
            "snapshots": parsed_args.snapshots,
            "gigabytes": parsed_args.gigabytes,
            "snapshot_gigabytes": parsed_args.snapshot_gigabytes,
            "share_networks": parsed_args.share_networks,
            "per_share_gigabytes": parsed_args.per_share_gigabytes,
        }

        if parsed_args.share_type is not None:
            if share_client.api_version < api_versions.APIVersion('2.39'):
                raise exceptions.CommandError(_(
                    "'share type' quotas are available only starting with "
                    "'2.39' API microversion."))
            kwargs["share_type"] = parsed_args.share_type
        if parsed_args.share_groups is not None:
            if share_client.api_version < api_versions.APIVersion('2.40'):
                raise exceptions.CommandError(_(
                    "'share group' quotas are available only starting with "
                    "'2.40' API microversion."))
            kwargs["share_groups"] = parsed_args.share_groups
        if parsed_args.share_group_snapshots is not None:
            if share_client.api_version < api_versions.APIVersion('2.40'):
                raise exceptions.CommandError(_(
                    "'share group snapshots' quotas are available only "
                    "starting with '2.40' API microversion."))
            kwargs["share_group_snapshots"] = parsed_args.share_group_snapshots
        if parsed_args.share_replicas is not None:
            if share_client.api_version < api_versions.APIVersion('2.53'):
                raise exceptions.CommandError(_(
                    "setting the number of 'share replicas' is available only "
                    "starting with API microversion '2.53'."))
            kwargs["share_replicas"] = parsed_args.share_replicas
        if parsed_args.replica_gigabytes is not None:
            if share_client.api_version < api_versions.APIVersion('2.53'):
                raise exceptions.CommandError(_(
                    "setting the capacity of share replicas in total "
                    "is available only starting with API microversion '2.53'.")
                )
            kwargs["replica_gigabytes"] = parsed_args.replica_gigabytes
        if parsed_args.per_share_gigabytes is not None:
            if share_client.api_version < api_versions.APIVersion('2.62'):
                raise exceptions.CommandError(_(
                    "'per share gigabytes' quotas are available only "
                    "starting with '2.62' API microversion.")
                )
            kwargs["per_share_gigabytes"] = parsed_args.per_share_gigabytes

        if all(value is None for value in kwargs.values()):
            raise exceptions.CommandError(_(
                "Nothing to set. "
                "New quota must be specified to at least one of the following "
                "resources: 'shares', 'snapshots', 'gigabytes', "
                "'snapshot-gigabytes', 'share-networks', 'share-type', "
                "'share-groups', 'share-group-snapshots', 'share-replicas', "
                "'replica-gigabytes', 'per-share-gigabytes'"))

        if parsed_args.quota_class:
            kwargs.update({
                "class_name": parsed_args.project,
            })
            try:
                share_client.quota_classes.update(**kwargs)
            except Exception as e:
                raise exceptions.CommandError(_(
                    "Failed to set quotas for %s class: '%s'")
                    % (parsed_args.project, e))
        else:
            project_id = utils.find_resource(
                identity_client.projects,
                parsed_args.project).id

            kwargs.update({
                "tenant_id": project_id,
                "force": parsed_args.force,
                "user_id": user_id
            })

            try:
                share_client.quotas.update(**kwargs)
            except Exception as e:
                raise exceptions.CommandError(_(
                    "Failed to set quotas for project '%s' : '%s'")
                    % (parsed_args.project, e))


class QuotaShow(command.ShowOne):
    """List the quotas for a project or project/user or project/share-type."""
    _description = _("Show Quota")

    def get_parser(self, prog_name):
        parser = super(QuotaShow, self).get_parser(prog_name)
        quota_type = parser.add_mutually_exclusive_group()
        parser.add_argument(
            'project',
            metavar='<project>',
            help=_('Name or ID of the project to list quotas for.')
        )
        quota_type.add_argument(
            '--user',
            metavar='<user>',
            default=None,
            help=_("Name or ID of user to list the quotas for. Optional. "
                   "Mutually exclusive with '--share-type'.")
        )
        quota_type.add_argument(
            '--share-type',
            metavar='<share-type>',
            type=str,
            default=None,
            help=_("Name or ID of a share type to list the quotas for. "
                   "Optional. "
                   "Mutually exclusive with '--user'. "
                   "Available only for microversion >= 2.39")
        )
        parser.add_argument(
            '--detail',
            action='store_true',
            default=False,
            help=_('Optional flag to indicate whether to show quota in detail.'
                   ' Default false, available only for microversion >= 2.25.')
        )
        parser.add_argument(
            '--defaults',
            action='store_true',
            default=False,
            help=_('Show the default quotas for the project.')
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        identity_client = self.app.client_manager.identity

        user_id = None
        if parsed_args.user:
            user_id = utils.find_resource(
                identity_client.users,
                parsed_args.user).id

        project_id = utils.find_resource(
            identity_client.projects,
            parsed_args.project).id

        quotas = {}
        if parsed_args.defaults:
            quotas = share_client.quotas.defaults(project_id)
        else:
            kwargs = {
                "tenant_id": project_id,
                "user_id": user_id,
                "detail": parsed_args.detail,
            }
            if parsed_args.share_type is not None:
                if share_client.api_version < api_versions.APIVersion("2.39"):
                    raise exceptions.CommandError(_(
                        "'share type' quotas are available only starting with "
                        "'2.39' API microversion."))
                kwargs["share_type"] = parsed_args.share_type

            quotas = share_client.quotas.get(**kwargs)

        printable_quotas = {}
        for quota_k, quota_v in sorted(quotas.to_dict().items()):
            if isinstance(quota_v, dict):
                quota_v = '\n'.join(
                    ['%s = %s' % (k, v) for k, v in sorted(quota_v.items())])
            printable_quotas[quota_k] = quota_v

        return self.dict2columns(printable_quotas)


class QuotaDelete(command.Command):
    """Delete quota for project/user or project/share-type.

    The quota will revert back to default.
    """

    _description = _("Delete Quota")

    def get_parser(self, prog_name):
        parser = super(QuotaDelete, self).get_parser(prog_name)
        quota_type = parser.add_mutually_exclusive_group()
        parser.add_argument(
            'project',
            metavar='<project>',
            help=_('Name or ID of the project to delete quotas for.')
        )
        quota_type.add_argument(
            '--user',
            metavar='<user>',
            default=None,
            help=_("Name or ID of user to delete the quotas for. Optional. "
                   "Mutually exclusive with '--share-type'.")
        )
        quota_type.add_argument(
            '--share-type',
            metavar='<share-type>',
            type=str,
            default=None,
            help=_("Name or ID of a share type to delete the quotas for. "
                   "Optional. "
                   "Mutually exclusive with '--user'. "
                   "Available only for microversion >= 2.39")
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        identity_client = self.app.client_manager.identity

        user_id = None
        if parsed_args.user:
            user_id = utils.find_resource(
                identity_client.users,
                parsed_args.user).id

        project_id = utils.find_resource(
            identity_client.projects,
            parsed_args.project).id

        kwargs = {
            "tenant_id": project_id,
            "user_id": user_id
        }
        if parsed_args.share_type:
            if share_client.api_version < api_versions.APIVersion("2.39"):
                raise exceptions.CommandError(_(
                    "'share type' quotas are available only starting with "
                    "API microversion '2.39'."))
            kwargs["share_type"] = parsed_args.share_type

        share_client.quotas.delete(**kwargs)



# Copyright 2015 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
from oslo_utils import uuidutils

from manilaclient.tests.functional import base


@ddt.ddt
class ExportLocationReadWriteTest(base.BaseTestCase):

    def setUp(self):
        super(ExportLocationReadWriteTest, self).setUp()
        self.share = self.create_share(
            client=self.get_user_client())

    @ddt.data('admin', 'user')
    def test_list_share_export_locations(self, role):
        self.skip_if_microversion_not_supported('2.14')

        client = self.admin_client if role == 'admin' else self.user_client
        export_locations = client.list_share_export_locations(
            self.share['id'])

        self.assertGreater(len(export_locations), 0)
        expected_keys = ('ID', 'Path', 'Preferred')
        for el in export_locations:
            for key in expected_keys:
                self.assertIn(key, el)
            self.assertTrue(uuidutils.is_uuid_like(el['ID']))
            self.assertIn(el['Preferred'], ('True', 'False'))

    @ddt.data('admin', 'user')
    def test_list_share_export_locations_with_columns(self, role):
        self.skip_if_microversion_not_supported('2.9')

        client = self.admin_client if role == 'admin' else self.user_client
        export_locations = client.list_share_export_locations(
            self.share['id'], columns='id,path')

        self.assertGreater(len(export_locations), 0)
        expected_keys = ('Id', 'Path')
        unexpected_keys = ('Updated At', 'Created At')
        for el in export_locations:
            for key in expected_keys:
                self.assertIn(key, el)
            for key in unexpected_keys:
                self.assertNotIn(key, el)
            self.assertTrue(uuidutils.is_uuid_like(el['Id']))

    @ddt.data('admin', 'user')
    def test_get_share_export_location(self, role):
        self.skip_if_microversion_not_supported('2.14')

        client = self.admin_client if role == 'admin' else self.user_client
        export_locations = client.list_share_export_locations(
            self.share['id'])

        el = client.get_share_export_location(
            self.share['id'], export_locations[0]['ID'])

        expected_keys = ['path', 'updated_at', 'created_at', 'id', 'preferred']
        if role == 'admin':
            expected_keys.extend(['is_admin_only', 'share_instance_id'])
        for key in expected_keys:
            self.assertIn(key, el)
        if role == 'admin':
            self.assertTrue(uuidutils.is_uuid_like(el['share_instance_id']))
            self.assertIn(el['is_admin_only'], ('True', 'False'))
        self.assertTrue(uuidutils.is_uuid_like(el['id']))
        self.assertIn(el['preferred'], ('True', 'False'))
        for list_k, get_k in (
                ('ID', 'id'), ('Path', 'path'), ('Preferred', 'preferred')):
            self.assertEqual(
                export_locations[0][list_k], el[get_k])

    def test_list_share_instance_export_locations(self):
        self.skip_if_microversion_not_supported('2.14')

        client = self.admin_client
        share_instances = client.list_share_instances(self.share['id'])
        self.assertGreater(len(share_instances), 0)
        self.assertIn('ID', share_instances[0])
        self.assertTrue(uuidutils.is_uuid_like(share_instances[0]['ID']))
        share_instance_id = share_instances[0]['ID']

        export_locations = client.list_share_instance_export_locations(
            share_instance_id)

        self.assertGreater(len(export_locations), 0)
        expected_keys = ('ID', 'Path', 'Is Admin only', 'Preferred')
        for el in export_locations:
            for key in expected_keys:
                self.assertIn(key, el)
            self.assertTrue(uuidutils.is_uuid_like(el['ID']))

    def test_list_share_instance_export_locations_with_columns(self):
        self.skip_if_microversion_not_supported('2.9')

        client = self.admin_client
        share_instances = client.list_share_instances(self.share['id'])
        self.assertGreater(len(share_instances), 0)
        self.assertIn('ID', share_instances[0])
        self.assertTrue(uuidutils.is_uuid_like(share_instances[0]['ID']))
        share_instance_id = share_instances[0]['ID']

        export_locations = client.list_share_instance_export_locations(
            share_instance_id, columns='id,path')

        self.assertGreater(len(export_locations), 0)
        expected_keys = ('Id', 'Path')
        unexpected_keys = ('Updated At', 'Created At', 'Is Admin only')
        for el in export_locations:
            for key in expected_keys:
                self.assertIn(key, el)
            for key in unexpected_keys:
                self.assertNotIn(key, el)
            self.assertTrue(uuidutils.is_uuid_like(el['Id']))

    def test_get_share_instance_export_location(self):
        self.skip_if_microversion_not_supported('2.14')

        client = self.admin_client
        share_instances = client.list_share_instances(self.share['id'])
        self.assertGreater(len(share_instances), 0)
        self.assertIn('ID', share_instances[0])
        self.assertTrue(uuidutils.is_uuid_like(share_instances[0]['ID']))
        share_instance_id = share_instances[0]['ID']

        export_locations = client.list_share_instance_export_locations(
            share_instance_id)

        el = client.get_share_instance_export_location(
            share_instance_id, export_locations[0]['ID'])

        expected_keys = (
            'path', 'updated_at', 'created_at', 'id', 'preferred',
            'is_admin_only', 'share_instance_id',
        )
        for key in expected_keys:
            self.assertIn(key, el)
        self.assertIn(el['is_admin_only'], ('True', 'False'))
        self.assertIn(el['preferred'], ('True', 'False'))
        self.assertTrue(uuidutils.is_uuid_like(el['id']))
        for list_k, get_k in (
                ('ID', 'id'), ('Path', 'path'), ('Preferred', 'preferred'),
                ('Is Admin only', 'is_admin_only')):
            self.assertEqual(
                export_locations[0][list_k], el[get_k])



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import json
import time

from tempest.lib.cli import base
from tempest.lib.common.utils import data_utils

from manilaclient import config

CONF = config.CONF


class OSCClientTestBase(base.ClientTestBase):
    """Base class for OSC manila functional tests"""

    @classmethod
    def get_admin_client(cls):
        admin_client = base.CLIClient(
            username=CONF.admin_username,
            password=CONF.admin_password,
            tenant_name=CONF.admin_tenant_name,
            uri=CONF.admin_auth_url,
            cli_dir=CONF.manila_exec_dir,
            insecure=CONF.insecure,
            project_domain_name=CONF.admin_project_domain_name or None,
            project_domain_id=CONF.admin_project_domain_id or None,
            user_domain_name=CONF.admin_user_domain_name or None,
            user_domain_id=CONF.admin_user_domain_id or None
        )
        return admin_client

    @classmethod
    def get_user_client(cls):
        user_client = base.CLIClient(
            username=CONF.username,
            password=CONF.password,
            tenant_name=CONF.tenant_name,
            uri=CONF.auth_url,
            cli_dir=CONF.manila_exec_dir,
            insecure=CONF.insecure,
            project_domain_name=CONF.project_domain_name or None,
            project_domain_id=CONF.project_domain_id or None,
            user_domain_name=CONF.user_domain_name or None,
            user_domain_id=CONF.user_domain_id or None
        )
        return user_client

    @property
    def admin_client(self):
        if not hasattr(self, '_admin_client'):
            self._admin_client = self.get_admin_client()
        return self._admin_client

    @property
    def user_client(self):
        if not hasattr(self, '_user_client'):
            self._user_client = self.get_user_client()
        return self._user_client

    def _get_clients(self):
        return self.admin_client

    def _get_property_from_output(self, output):
        """Creates a dictionary from the given output"""
        obj = {}
        items = self.parser.listing(output)
        for item in items:
            obj[item['Field']] = str(item['Value'])
        return obj

    def _wait_for_object_status(self, object_name, object_id, status,
                                timeout=CONF.build_timeout,
                                interval=CONF.build_interval):
        """Waits for a object to reach a given status."""

        start_time = time.time()
        while time.time() - start_time < timeout:
            if status == self.openstack(
                    '%(obj)s show -c status -f value %(id)s'
                    % {'obj': object_name,
                       'id': object_id}).rstrip():
                break
            time.sleep(interval)
        else:
            self.fail("%s %s did not reach status %s after %d seconds."
                      % (object_name, object_id, status, timeout))

    def check_object_deleted(self, object_name, object_id,
                             timeout=CONF.build_timeout,
                             interval=CONF.build_interval):
        """Check that object deleted successfully"""

        cmd = '%s list -c ID -f value' % object_name
        start_time = time.time()
        while time.time() - start_time < timeout:
            if object_id not in self.openstack(cmd):
                break
            time.sleep(interval)
        else:
            self.fail("%s %s not deleted after %d seconds."
                      % (object_name, object_id, timeout))

    def openstack(self, action, flags='', params='', fail_ok=False,
                  merge_stderr=False, client=None):
        """Executes openstack command for given action"""

        if '--os-share-api-version' not in flags:
            flags = (
                flags + '--os-share-api-version %s'
                % CONF.max_api_microversion)

        if client is None:
            client = self.admin_client

        return client.openstack(action, flags=flags, params=params,
                                fail_ok=fail_ok,
                                merge_stderr=merge_stderr)

    def listing_result(self, object_name, command, client=None):
        """Returns output for the given command as list of dictionaries"""

        output = self.openstack(object_name, params=command, client=client)
        result = self.parser.listing(output)
        return result

    def dict_result(self, object_name, command, client=None):
        """Returns output for the given command as dictionary"""

        output = self.openstack(object_name, params=command, client=client)
        result_dict = self._get_property_from_output(output)
        return result_dict

    def create_share(self, share_protocol=None, size=None, name=None,
                     snapshot_id=None, properties=None, share_network=None,
                     description=None, public=False, share_type=None,
                     availability_zone=None, share_group=None,
                     add_cleanup=True, client=None,
                     wait_for_status='available'):

        name = name or data_utils.rand_name('autotest_share_name')
        # share_type = dhss_false until we have implemented
        # share network commands for osc
        share_type = share_type or 'dhss_false'

        cmd = ('create '
               '%(protocol)s %(size)s %(name)s %(desc)s %(public)s %(stype)s'
               % {'protocol': share_protocol or 'NFS',
                  'size': size or '1',
                  'name': '--name %s' % name,
                  'desc': '--description %s' % description,
                  'public': '--public %s' % public,
                  'stype': '--share-type %s' % share_type})

        if snapshot_id:
            cmd = cmd + ' --snapshot-id %s' % snapshot_id
        if properties:
            for key, value in properties.items():
                cmd = (cmd + ' --property %(key)s=%(value)s'
                       % {'key': key, 'value': value})
        if share_network:
            cmd = cmd + ' --share-network %s' % share_network
        if availability_zone:
            cmd = cmd + ' --availability-zone %s' % availability_zone
        if share_group:
            cmd = cmd + ' --share-group %s' % share_group

        share_object = self.dict_result('share', cmd, client=client)
        self._wait_for_object_status(
            'share', share_object['id'], wait_for_status)

        if add_cleanup:
            self.addCleanup(
                self.openstack, 'share delete %s --wait' % share_object['id']
            )
        return share_object

    def list_pools(self, backend=None, host=None, pool=None, detail=False):
        cmd = 'pool list '

        if backend:
            cmd += f'--backend {backend} '
        if pool:
            cmd += f'--pool {pool} '
        if host:
            cmd += f'--host {host} '
        if detail:
            cmd += '--detail'

        pools = self.listing_result('share', cmd)

        return pools

    def create_share_type(self, name=None, dhss=False, description=None,
                          snapshot_support=None,
                          create_share_from_snapshot_support=None,
                          revert_to_snapshot_support=False,
                          mount_snapshot_support=False, extra_specs={},
                          public=True, add_cleanup=True, client=None,
                          formatter=None):

        name = name or data_utils.rand_name('autotest_share_type_name')

        cmd = (f'create {name} {dhss} --public {public}')
        if description:
            cmd += f' --description {description}'
        if snapshot_support:
            cmd += f' --snapshot-support {snapshot_support}'
        if create_share_from_snapshot_support:
            cmd += (' --create-share-from-snapshot-support '
                    f'{create_share_from_snapshot_support}')
        if revert_to_snapshot_support:
            cmd += (' --revert-to-snapshot-support '
                    f' {revert_to_snapshot_support}')
        if mount_snapshot_support:
            cmd += f' --mount-snapshot-support {mount_snapshot_support}'
        if extra_specs:
            specs = ' --extra-specs'
            for key, value in extra_specs.items():
                specs += f' {key}={value}'
            cmd += specs

        if formatter == 'json':
            cmd = f'share type {cmd} -f {formatter} '
            share_type = json.loads(self.openstack(cmd, client=client))
        else:
            share_type = self.dict_result('share type', cmd, client=client)

        if add_cleanup:
            self.addCleanup(
                self.openstack, f'share type delete {share_type["id"]}'
            )
        return share_type

    def list_services(self, host=None, status=None, state=None, zone=None):
        cmd = 'service list '

        if host:
            cmd += f'--host {host} '
        if status:
            cmd += f'--status {status} '
        if state:
            cmd += f'--state {state} '
        if zone:
            cmd += f'--zone {zone} '

        services = self.listing_result('share', cmd)
        return services

    def create_share_access_rule(self, share, access_type,
                                 access_to, properties=None,
                                 access_level=None, wait=False):
        cmd = f'access create {share} {access_type} {access_to} '

        if access_level:
            cmd += f'--access-level {access_level} '
        if properties:
            cmd += f'--properties {properties} '
        if wait:
            cmd += f'--wait'

        access_rule = self.dict_result('share', cmd)

        return access_rule

    def get_share_export_locations(self, share):
        cmd = (f'export location list {share}')
        export_locations = json.loads(self.openstack(f'share {cmd} -f json'))
        return export_locations

    def create_snapshot(self, share, name=None,
                        description=None, wait=True, force=None,
                        add_cleanup=True, client=None):

        name = name or data_utils.rand_name('autotest_snapshot_name')

        cmd = (f'snapshot create {share} --name {name} ')

        if description:
            cmd += f' --description {description}'
        if wait:
            cmd += ' --wait'
        if force:
            cmd += ' --force'

        snapshot_object = self.dict_result('share', cmd, client=client)

        if add_cleanup:
            self.addCleanup(
                self.openstack,
                f'share snapshot delete {snapshot_object["id"]} --wait')

        return snapshot_object

    def create_share_network(self, neutron_net_id=None,
                             neutron_subnet_id=None, name=None,
                             description=None, availability_zone=None,
                             add_cleanup=True):
        name = name or data_utils.rand_name('autotest_share_network_name')
        cmd = (f'network create --name {name} --description {description}')
        if neutron_net_id:
            cmd = cmd + f' --neutron-net-id {neutron_net_id}'
        if neutron_subnet_id:
            cmd = cmd + f' --neutron-subnet-id {neutron_subnet_id}'
        if availability_zone:
            cmd = cmd + f' --availability-zone {availability_zone}'

        share_network_obj = self.dict_result('share', cmd)
        self._wait_for_object_status(
            'share network', share_network_obj['id'], 'active')
        if add_cleanup:
            self.addCleanup(
                self.openstack,
                f'share network delete {share_network_obj["id"]}'
            )
        return share_network_obj

    def create_share_replica(self, share, availability_zone=None,
                             share_network=None, wait=None,
                             add_cleanup=True):
        cmd = (f'replica create {share}')

        if availability_zone:
            cmd = cmd + f' --availability-zone {availability_zone}'
        if wait:
            cmd = cmd + ' --wait'
        if share_network:
            cmd = cmd + ' --share-network %s' % share_network

        replica_object = self.dict_result('share', cmd)
        self._wait_for_object_status(
            'share replica', replica_object['id'], 'available')

        if add_cleanup:
            self.addCleanup(
                self.openstack,
                f'share replica delete {replica_object["id"]} --wait'
            )
        return replica_object

    def get_share_replica_export_locations(self, replica):
        cmd = (f'replica export location list {replica}')
        export_locations = self.listing_result('share', cmd)
        return export_locations

    def create_share_group_type(self, name=None, share_types=None,
                                group_specs=None, public=True,
                                add_cleanup=True):

        name = name or data_utils.rand_name('autotest_share_group_types_name')
        share_types = share_types or 'None'

        cmd = (f'group type create '
               f'{name} '
               f'{share_types} ')

        if group_specs:
            cmd = cmd + f' --group-specs {group_specs} '
        if not public:
            cmd = cmd + f' --public {public} '

        share_object = self.dict_result('share', cmd)

        if add_cleanup:
            self.addCleanup(
                self.openstack,
                'share group type delete %s' % share_object['id'])
        return share_object

    def share_group_type_access_create(self, group_type, project):
        cmd = (f'group type access create '
               f'{group_type} '
               f'{project} ')

        self.dict_result('share', cmd)

    def share_group_type_access_delete(self, group_type, access_id):
        cmd = (f'group type access delete '
               f'{group_type} '
               f'{access_id} ')

        self.dict_result('share', cmd)

    def check_create_network_subnet(self, share_network, neutron_net_id=None,
                                    neutron_subnet_id=None,
                                    availability_zone=None,
                                    restart_check=None):
        cmd = f'network subnet create {share_network} --check-only'

        if neutron_net_id:
            cmd += f' --neutron-net-id {neutron_net_id}'
        if neutron_subnet_id:
            cmd += f' --neutron-subnet-id {neutron_subnet_id}'
        if availability_zone:
            cmd += f' --availability-zone {availability_zone}'
        if restart_check:
            cmd += f' --restart-check'

        check_result = self.dict_result('share', cmd)
        return check_result



# Copyright 2010 Jacob Kaplan-Moss

# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import ddt

from manilaclient import api_versions
from manilaclient.common.apiclient import exceptions as client_exceptions
from manilaclient import exceptions
from manilaclient import extension
from manilaclient.tests.unit import utils
from manilaclient.tests.unit.v2 import fakes
from manilaclient.v2 import shares

extensions = [
    extension.Extension('shares', shares),
]
cs = fakes.FakeClient(extensions=extensions)


@ddt.ddt
class SharesTest(utils.TestCase):

    # Testcases for class Share
    def setUp(self):
        super(SharesTest, self).setUp()
        self.share = shares.Share(None, {'id': 1})
        self.share.manager = mock.Mock()

    def test_share_allow_access(self):
        access_level = 'fake_level'
        access_to = 'fake_credential'
        access_type = 'fake_type'

        self.share.allow(access_type, access_to, access_level)

        self.share.manager.allow.assert_called_once_with(
            self.share, access_type, access_to, access_level, None)

    # Testcases for class ShareManager

    @ddt.data('nfs', 'cifs', 'cephfs', 'glusterfs', 'hdfs', 'maprfs')
    def test_create_share_with_protocol(self, protocol):
        expected = {
            'size': 1,
            'snapshot_id': None,
            'name': None,
            'description': None,
            'metadata': dict(),
            'share_proto': protocol,
            'share_network_id': None,
            'share_type': None,
            'is_public': False,
            'availability_zone': None,
            'scheduler_hints': dict(),
        }
        cs.shares.create(protocol, 1)
        cs.assert_called('POST', '/shares', {'share': expected})

    @ddt.data(
        type('ShareNetworkUUID', (object, ), {'uuid': 'fake_nw'}),
        type('ShareNetworkID', (object, ), {'id': 'fake_nw'}),
        'fake_nw')
    def test_create_share_with_share_network(self, share_network):
        expected = {
            'size': 1,
            'snapshot_id': None,
            'name': None,
            'description': None,
            'metadata': dict(),
            'share_proto': 'nfs',
            'share_network_id': 'fake_nw',
            'share_type': None,
            'is_public': False,
            'availability_zone': None,
            'scheduler_hints': dict(),
        }
        cs.shares.create('nfs', 1, share_network=share_network)
        cs.assert_called('POST', '/shares', {'share': expected})

    @ddt.data(
        type('ShareTypeUUID', (object, ), {'uuid': 'fake_st'}),
        type('ShareTypeID', (object, ), {'id': 'fake_st'}),
        'fake_st')
    def test_create_share_with_share_type(self, share_type):
        expected = {
            'size': 1,
            'snapshot_id': None,
            'name': None,
            'description': None,
            'metadata': dict(),
            'share_proto': 'nfs',
            'share_network_id': None,
            'share_type': 'fake_st',
            'is_public': False,
            'availability_zone': None,
            'scheduler_hints': dict(),
        }
        cs.shares.create('nfs', 1, share_type=share_type)
        cs.assert_called('POST', '/shares', {'share': expected})

    @ddt.data({'is_public': True,
               'availability_zone': 'nova'},
              {'is_public': False,
               'availability_zone': 'fake_azzzzz'})
    @ddt.unpack
    def test_create_share_with_all_params_defined(self, is_public,
                                                  availability_zone):
        body = {
            'share': {
                'is_public': is_public,
                'share_type': None,
                'name': None,
                'snapshot_id': None,
                'description': None,
                'metadata': {},
                'share_proto': 'nfs',
                'share_network_id': None,
                'size': 1,
                'availability_zone': availability_zone,
                'scheduler_hints': {},
            }
        }
        cs.shares.create('nfs', 1, is_public=is_public,
                         availability_zone=availability_zone)
        cs.assert_called('POST', '/shares', body)

    @ddt.data(
        type('ShareUUID', (object, ), {'uuid': '1234'}),
        type('ShareID', (object, ), {'id': '1234'}),
        '1234')
    def test_get_share(self, share):
        share = cs.shares.get(share)
        cs.assert_called('GET', '/shares/1234')

    @ddt.data(
        type('ShareUUID', (object, ), {'uuid': '1234'}),
        type('ShareID', (object, ), {'id': '1234'}),
        '1234')
    def test_get_update(self, share):
        data = dict(foo='bar', quuz='foobar')
        share = cs.shares.update(share, **data)
        cs.assert_called('PUT', '/shares/1234', {'share': data})

    def test_delete_share(self):
        share = cs.shares.get('1234')
        cs.shares.delete(share)
        cs.assert_called('DELETE', '/shares/1234')

    def test_soft_delete_share(self):
        share = cs.shares.get('1234')
        cs.shares.soft_delete(share)
        body = {"soft_delete": None}
        cs.assert_called('POST', '/shares/1234/action', body=body)

    def test_restore_share(self):
        share = cs.shares.get('1234')
        cs.shares.restore(share)
        body = {"restore": None}
        cs.assert_called('POST', '/shares/1234/action', body=body)

    @ddt.data(
        ("2.6", "/os-share-manage", None),
        ("2.7", "/shares/manage", None),
        ("2.8", "/shares/manage", True),
        ("2.8", "/shares/manage", False),
        ("2.49", "/shares/manage", False, '1234'),
    )
    @ddt.unpack
    def test_manage_share(self, microversion, resource_path, is_public=False,
                          share_server_id=None):
        service_host = "fake_service_host"
        protocol = "fake_protocol"
        export_path = "fake_export_path"
        driver_options = "fake_driver_options"
        share_type = "fake_share_type"
        name = "foo_name"
        description = "bar_description"
        expected_body = {
            "service_host": service_host,
            "share_type": share_type,
            "protocol": protocol,
            "export_path": export_path,
            "driver_options": driver_options,
            "name": name,
            "description": description,
            "share_server_id": share_server_id,
        }
        version = api_versions.APIVersion(microversion)
        if version >= api_versions.APIVersion('2.8'):
            expected_body["is_public"] = is_public

        mock_microversion = mock.Mock(api_version=version)
        manager = shares.ShareManager(api=mock_microversion)

        with mock.patch.object(manager, "_create",
                               mock.Mock(return_value="fake")):
            if version < api_versions.APIVersion('2.8'):
                result = manager.manage(
                    service_host, protocol, export_path, driver_options,
                    share_type, name, description)
            elif (api_versions.APIVersion('2.8') <= version
                    < api_versions.APIVersion('2.49')):
                result = manager.manage(
                    service_host, protocol, export_path, driver_options,
                    share_type, name, description, is_public)
            else:
                result = manager.manage(
                    service_host, protocol, export_path, driver_options,
                    share_type, name, description, is_public, share_server_id)

            self.assertEqual(manager._create.return_value, result)
            manager._create.assert_called_once_with(
                resource_path, {"share": expected_body}, "share")

    @ddt.data(
        type("ShareUUID", (object, ), {"uuid": "1234"}),
        type("ShareID", (object, ), {"id": "1234"}),
        "1234")
    def test_unmanage_share_v2_6(self, share):
        version = api_versions.APIVersion("2.6")
        mock_microversion = mock.Mock(api_version=version)
        manager = shares.ShareManager(api=mock_microversion)

        with mock.patch.object(manager, "_action",
                               mock.Mock(return_value="fake")):
            result = manager.unmanage(share)

            self.assertFalse(manager._action.called)
            self.assertNotEqual("fake", result)
            self.assertEqual(manager.api.client.post.return_value, result)
            manager.api.client.post.assert_called_once_with(
                "/os-share-unmanage/1234/unmanage")

    def test_unmanage_share_v2_7(self):
        share = "fake_share"
        version = api_versions.APIVersion("2.7")
        mock_microversion = mock.Mock(api_version=version)
        manager = shares.ShareManager(api=mock_microversion)

        with mock.patch.object(manager, "_action",
                               mock.Mock(return_value="fake")):
            result = manager.unmanage(share)

            manager._action.assert_called_once_with("unmanage", share)
            self.assertEqual("fake", result)

    def test_revert_to_snapshot(self):

        share = 'fake_share'
        snapshot = 'fake_snapshot'
        version = api_versions.APIVersion("2.27")
        mock_microversion = mock.Mock(api_version=version)
        manager = shares.ShareManager(api=mock_microversion)
        mock_action = self.mock_object(
            manager, '_action', mock.Mock(return_value='fake'))

        result = manager.revert_to_snapshot(share, snapshot)

        self.assertEqual('fake', result)
        mock_action.assert_called_once_with(
            'revert', 'fake_share', info={'snapshot_id': 'fake_snapshot'})

    def test_revert_to_snapshot_not_supported(self):

        share = 'fake_share'
        snapshot = 'fake_snapshot'
        version = api_versions.APIVersion("2.26")
        mock_microversion = mock.Mock(api_version=version)
        manager = shares.ShareManager(api=mock_microversion)

        self.assertRaises(client_exceptions.UnsupportedVersion,
                          manager.revert_to_snapshot,
                          share,
                          snapshot)

    @ddt.data(
        ("2.6", "os-force_delete"),
        ("2.7", "force_delete"),
    )
    @ddt.unpack
    def test_force_delete_share(self, microversion, action_name):
        share = "fake_share"
        version = api_versions.APIVersion(microversion)
        mock_microversion = mock.Mock(api_version=version)
        manager = shares.ShareManager(api=mock_microversion)

        with mock.patch.object(manager, "_action",
                               mock.Mock(return_value="fake")):
            result = manager.force_delete(share)

            manager._action.assert_called_once_with(action_name, share)
            self.assertEqual("fake", result)

    def test_list_shares_index(self):
        cs.shares.list(detailed=False)
        cs.assert_called('GET', '/shares?is_public=True')

    @ddt.data("1.0", "2.35", "2.69")
    def test_list_shares_index_diff_api_version(self, microversion):
        version = api_versions.APIVersion(microversion)
        mock_microversion = mock.Mock(api_version=version)
        manager = shares.ShareManager(api=mock_microversion)
        search_opts1 = {}
        search_opts2 = {
            'export_location': 'fake_export_id',
        }
        search_opts3 = {
            'export_location': 'fake_export_id',
            'is_soft_deleted': 'True',
        }

        with mock.patch.object(manager, "do_list",
                               mock.Mock(return_value="fake")):
            manager.list(detailed=False, search_opts=search_opts3)

            if version >= api_versions.APIVersion('2.69'):
                manager.do_list.assert_called_once_with(
                    detailed=False, search_opts=search_opts3,
                    sort_key=None, sort_dir=None)
            elif version >= api_versions.APIVersion('2.35'):
                manager.do_list.assert_called_once_with(
                    detailed=False, search_opts=search_opts2,
                    sort_key=None, sort_dir=None)
            else:
                manager.do_list.assert_called_once_with(
                    detailed=False, search_opts=search_opts1,
                    sort_key=None, sort_dir=None)

    def test_list_shares_index_with_search_opts(self):
        search_opts = {
            'fake_str': 'fake_str_value',
            'fake_int': 1,
            'name~': 'fake_name',
            'description~': 'fake_description',
        }
        cs.shares.list(detailed=False, search_opts=search_opts)
        cs.assert_called(
            'GET',
            '/shares?description~=fake_description&fake_int=1&'
            'fake_str=fake_str_value&is_public=True&name~=fake_name')

    @ddt.data(('id', 'b4991315-eb7d-43ec-979e-5715d4399827', True),
              ('id', 'b4991315-eb7d-43ec-979e-5715d4399827', False),
              ('path', 'fake_path', False),
              ('path', 'fake_path', True))
    @ddt.unpack
    def test_list_shares_index_with_export_location(self, filter_type,
                                                    value, detailed):
        search_opts = {
            'export_location': value,
        }
        cs.shares.list(detailed=detailed, search_opts=search_opts)
        if detailed:
            cs.assert_called(
                'GET', ('/shares/detail?export_location_' + filter_type +
                        '=' + value + '&is_public=True'))
        else:
            cs.assert_called(
                'GET', ('/shares?export_location_' + filter_type + '='
                        + value + '&is_public=True'))

    @ddt.data(True, False)
    def test_list_shares_index_with_is_soft_deleted(self, detailed):
        search_opts = {
            'is_soft_deleted': 'True',
        }
        cs.shares.list(detailed=detailed, search_opts=search_opts)
        if detailed:
            cs.assert_called(
                'GET', ('/shares/detail?is_public=True'
                        + '&is_soft_deleted=True'))
        else:
            cs.assert_called(
                'GET', ('/shares?is_public=True'
                        + '&is_soft_deleted=True'))

    def test_list_shares_detailed(self):
        search_opts = {
            'with_count': 'True',
        }
        shares, count = cs.shares.list(detailed=True, search_opts=search_opts)
        cs.assert_called(
            'GET', '/shares/detail?is_public=True&with_count=True')
        self.assertEqual(2, count)
        self.assertEqual(1, len(shares))

    def test_list_shares_detailed_with_count(self):
        cs.shares.list(detailed=True)
        cs.assert_called('GET', '/shares/detail?is_public=True')

    def test_list_shares_detailed_with_search_opts(self):
        search_opts = {
            'fake_str': 'fake_str_value',
            'fake_int': 1,
        }
        cs.shares.list(detailed=True, search_opts=search_opts)
        cs.assert_called(
            'GET',
            '/shares/detail?fake_int=1&fake_str=fake_str_value&is_public=True')

    def test_list_shares_sort_by_asc_and_host_key(self):
        cs.shares.list(detailed=False, sort_key='host', sort_dir='asc')
        cs.assert_called('GET',
                         '/shares?is_public=True&sort_dir=asc&sort_key=host')

    def test_list_shares_sort_by_desc_and_size_key(self):
        cs.shares.list(detailed=False, sort_key='size', sort_dir='desc')
        cs.assert_called('GET',
                         '/shares?is_public=True&sort_dir=desc&sort_key=size')

    def test_list_shares_filter_by_share_network_alias(self):
        cs.shares.list(detailed=False, sort_key='share_network')
        cs.assert_called('GET',
                         '/shares?is_public=True&sort_key=share_network_id')

    def test_list_shares_filter_by_snapshot_alias(self):
        cs.shares.list(detailed=False, sort_key='snapshot')
        cs.assert_called('GET', '/shares?is_public=True&sort_key=snapshot_id')

    def test_list_shares_filter_by_share_type_alias(self):
        cs.shares.list(detailed=False, sort_key='share_type')
        cs.assert_called('GET',
                         '/shares?is_public=True&sort_key=share_type_id')

    def test_list_shares_by_improper_direction(self):
        self.assertRaises(ValueError, cs.shares.list, sort_dir='fake')

    def test_list_shares_by_improper_key(self):
        self.assertRaises(ValueError, cs.shares.list, sort_key='fake')

    @ddt.data(
        {'access_to': '127.0.0.1', 'access_type': 'ip',
         'action_name': 'os-allow_access', 'microversion': '2.0'},
        {'access_to': '1' * 4, 'access_type': 'user',
         'action_name': 'os-allow_access', 'microversion': '2.0'},
        {'access_to': '1' * 255, 'access_type': 'user',
         'action_name': 'os-allow_access', 'microversion': '2.0'},
        {'access_to': 'fake${.-_\'`}', 'access_type': 'user',
         'action_name': 'os-allow_access', 'microversion': '2.0'},
        {'access_to': 'MYDOMAIN-Administrator', 'access_type': 'user',
         'action_name': 'os-allow_access', 'microversion': '2.0'},
        {'access_to': 'test group name', 'access_type': 'user',
         'action_name': 'os-allow_access', 'microversion': '2.0'},
        {'access_to': 'x', 'access_type': 'cert',
         'action_name': 'os-allow_access', 'microversion': '2.0'},
        {'access_to': 'x' * 64, 'access_type': 'cert',
         'action_name': 'os-allow_access', 'microversion': '2.0'},
        {'access_to': 'tenant.example.com', 'access_type': 'cert',
         'action_name': 'os-allow_access', 'microversion': '2.0'},
        {'access_to': '127.0.0.1', 'access_type': 'ip',
         'action_name': 'allow_access', 'microversion': '2.7'},
        {'access_to': 'test group name', 'access_type': 'user',
         'action_name': 'allow_access', 'microversion': '2.7'},
        {'access_to': 'alice', 'access_type': 'cephx',
         'action_name': 'allow_access', 'microversion': '2.13'},
        {'access_to': 'alice_bob', 'access_type': 'cephx',
         'action_name': 'allow_access', 'microversion': '2.13'},
        {'access_to': 'alice bob', 'access_type': 'cephx',
         'action_name': 'allow_access', 'microversion': '2.13'},
        {'access_to': 'test group name', 'access_type': 'user',
         'action_name': 'allow_access', 'microversion': '2.13'},
        {'access_to': 'AD80:0000:0000:0000:ABAA:0000:00C2:0002',
         'access_type': 'ip', 'action_name': 'allow_access',
         'microversion': '2.38'},
        {'access_to': 'AD80::/36',
         'access_type': 'ip', 'action_name': 'allow_access',
         'microversion': '2.38'},
        {'access_to': 'AD80:ABAA::/128',
         'access_type': 'ip', 'action_name': 'allow_access',
         'microversion': '2.38'},
        {'access_to': 'ad80::abaa:0:c2:2',
         'access_type': 'ip', 'action_name': 'allow_access',
         'microversion': '2.38'},
        {'access_to': 'test group name', 'access_type': 'user',
         'action_name': 'allow_access', 'microversion': '2.38'},
    )
    @ddt.unpack
    def test_allow_access_to_share(self, access_to, access_type,
                                   action_name, microversion):
        access = ('foo', {'access': 'bar'})
        access_level = 'fake_access_level'
        share = 'fake_share'
        version = api_versions.APIVersion(microversion)
        mock_microversion = mock.Mock(api_version=version)
        manager = shares.ShareManager(api=mock_microversion)

        with mock.patch.object(manager, '_action',
                               mock.Mock(return_value=access)):
            result = manager.allow(share, access_type, access_to, access_level)

            manager._action.assert_called_once_with(
                action_name, share, {'access_level': access_level,
                                     'access_type': access_type,
                                     'access_to': access_to})
            self.assertEqual('bar', result)

    @ddt.data(
        {'access_to': 'localhost', 'access_type': 'ip',
         'microversion': '2.0'},
        {'access_to': '127.0.0.*', 'access_type': 'ip',
         'microversion': '2.0'},
        {'access_to': '127.0.0.0/33', 'access_type': 'ip',
         'microversion': '2.0'},
        {'access_to': '127.0.0.256', 'access_type': 'ip',
         'microversion': '2.0'},
        {'access_to': '1', 'access_type': 'user',
         'microversion': '2.0'},
        {'access_to': '1' * 3, 'access_type': 'user',
         'microversion': '2.0'},
        {'access_to': '1' * 256, 'access_type': 'user',
         'microversion': '2.0'},
        {'access_to': 'root+=', 'access_type': 'user',
         'microversion': '2.0'},
        {'access_to': '', 'access_type': 'cert',
         'microversion': '2.0'},
        {'access_to': ' ', 'access_type': 'cert',
         'microversion': '2.0'},
        {'access_to': 'x' * 65, 'access_type': 'cert',
         'microversion': '2.0'},
        {'access_to': 'alice', 'access_type': 'cephx',
         'microversion': '2.0'},
        {'access_to': '', 'access_type': 'cephx',
         'microversion': '2.13'},
        {'access_to': u"bj\u00F6rn", 'access_type': 'cephx',
         'microversion': '2.13'},
        {'access_to': "AD80:0000:0000:0000:ABAA:0000:00C2:0002/65",
         'access_type': 'ip', 'microversion': '2.38'},
        {'access_to': "AD80:0000:0000:0000:ABAA:0000:00C2:0002*32",
         'access_type': 'ip', 'microversion': '2.38'},
        {'access_to': "AD80:0000:0000:0000:ABAA:0000:00C2:0002",
         'access_type': 'ip', 'microversion': '2.37'},
    )
    @ddt.unpack
    def test_allow_access_to_share_error_invalid_access(self, access_to,
                                                        access_type,
                                                        microversion):
        access = ('foo', {'access': 'bar'})
        access_level = 'fake_access_level'
        share = 'fake_share'
        version = api_versions.APIVersion(microversion)
        mock_microversion = mock.Mock(api_version=version)
        manager = shares.ShareManager(api=mock_microversion)

        with mock.patch.object(manager, '_action',
                               mock.Mock(return_value=access)):
            self.assertRaises(exceptions.CommandError, manager.allow,
                              share, access_type, access_to, access_level)

            manager._action.assert_not_called()

    @ddt.data(
        ("2.6", "os-deny_access"),
        ("2.7", "deny_access"),
    )
    @ddt.unpack
    def test_deny_access_to_share(self, microversion, action_name):
        access_id = "fake_access_id"
        share = "fake_share"
        version = api_versions.APIVersion(microversion)
        mock_microversion = mock.Mock(api_version=version)
        manager = shares.ShareManager(api=mock_microversion)

        with mock.patch.object(manager, "_action",
                               mock.Mock(return_value="fake")):
            result = manager.deny(share, access_id)

            manager._action.assert_called_once_with(
                action_name, share, {"access_id": access_id})
            self.assertEqual("fake", result)

    def test_get_metadata(self):
        cs.shares.get_metadata(1234)
        cs.assert_called('GET', '/shares/1234/metadata')

    def test_set_metadata(self):
        cs.shares.set_metadata(1234, {'k1': 'v2'})
        cs.assert_called('POST', '/shares/1234/metadata',
                         {'metadata': {'k1': 'v2'}})

    @ddt.data(
        type('ShareUUID', (object, ), {'uuid': '1234'}),
        type('ShareID', (object, ), {'id': '1234'}),
        '1234')
    def test_delete_metadata(self, share):
        keys = ['key1']
        cs.shares.delete_metadata(share, keys)
        cs.assert_called('DELETE', '/shares/1234/metadata/key1')

    @ddt.data(
        type('ShareUUID', (object, ), {'uuid': '1234'}),
        type('ShareID', (object, ), {'id': '1234'}),
        '1234')
    def test_metadata_update_all(self, share):
        cs.shares.update_all_metadata(share, {'k1': 'v1'})
        cs.assert_called('PUT', '/shares/1234/metadata',
                         {'metadata': {'k1': 'v1'}})

    @ddt.data(
        ("2.6", "os-reset_status"),
        ("2.7", "reset_status"),
    )
    @ddt.unpack
    def test_reset_share_state(self, microversion, action_name):
        state = "available"
        share = "fake_share"
        version = api_versions.APIVersion(microversion)
        mock_microversion = mock.Mock(api_version=version)
        manager = shares.ShareManager(api=mock_microversion)

        with mock.patch.object(manager, "_action",
                               mock.Mock(return_value="fake")):
            result = manager.reset_state(share, state)

            manager._action.assert_called_once_with(
                action_name, share, {"status": state})
            self.assertEqual("fake", result)

    @ddt.data(
        ("2.6", "os-extend", False),
        ("2.7", "extend", False),
        ("2.64", "extend", True),
    )
    @ddt.unpack
    def test_extend_share(self, microversion, action_name, force):
        size = 123
        share = "fake_share"
        version = api_versions.APIVersion(microversion)
        mock_microversion = mock.Mock(api_version=version)
        manager = shares.ShareManager(api=mock_microversion)

        with mock.patch.object(manager, "_action",
                               mock.Mock(return_value="fake")):
            if not force:
                result = manager.extend(share, size)
                manager._action.assert_called_once_with(
                    action_name, share, {"new_size": size})
            else:
                result = manager.extend(share, size, force=force)
                manager._action.assert_called_once_with(
                    action_name, share, {"new_size": size,
                                         "force": "true"})
            self.assertEqual("fake", result)

    @ddt.data(
        ("2.6", "os-shrink"),
        ("2.7", "shrink"),
    )
    @ddt.unpack
    def test_shrink_share(self, microversion, action_name):
        size = 123
        share = "fake_share"
        version = api_versions.APIVersion(microversion)
        mock_microversion = mock.Mock(api_version=version)
        manager = shares.ShareManager(api=mock_microversion)

        with mock.patch.object(manager, "_action",
                               mock.Mock(return_value="fake")):
            result = manager.shrink(share, size)

            manager._action.assert_called_once_with(
                action_name, share, {"new_size": size})
            self.assertEqual("fake", result)

    def test_list_share_instances(self):
        share = type('ShareID', (object, ), {'id': '1234'})
        cs.shares.list_instances(share)
        cs.assert_called('GET', '/shares/1234/instances')

    def test_migration_start(self):
        share = "fake_share"
        host = "fake_host"
        version = api_versions.APIVersion('2.29')
        manager = shares.ShareManager(
            api=fakes.FakeClient(api_version=version))

        with mock.patch.object(manager, "_action",
                               mock.Mock(return_value="fake")):
            result = manager.migration_start(
                share, host, force_host_assisted_migration=True,
                preserve_metadata=True, writable=True, nondisruptive=True,
                preserve_snapshots=True)
            manager._action.assert_called_once_with(
                'migration_start', share, {
                    "host": host,
                    "force_host_assisted_migration": True,
                    "preserve_metadata": True,
                    "writable": True,
                    "nondisruptive": True,
                    "preserve_snapshots": True,
                    "new_share_network_id": None,
                    "new_share_type_id": None,
                })

            self.assertEqual("fake", result)

    def test_migration_complete(self):
        share = "fake_share"
        manager = shares.ShareManager(api=fakes.FakeClient())

        with mock.patch.object(manager, "_action",
                               mock.Mock(return_value="fake")):
            result = manager.migration_complete(share)

            manager._action.assert_called_once_with(
                "migration_complete", share)
            self.assertEqual("fake", result)

    def test_migration_get_progress(self):
        share = "fake_share"
        manager = shares.ShareManager(api=fakes.FakeClient())

        with mock.patch.object(manager, "_action",
                               mock.Mock(return_value="fake")):
            result = manager.migration_get_progress(share)

            manager._action.assert_called_once_with(
                "migration_get_progress", share)
            self.assertEqual("fake", result)

    def test_reset_task_state(self):
        share = "fake_share"
        state = "fake_state"
        manager = shares.ShareManager(api=fakes.FakeClient())

        with mock.patch.object(manager, "_action",
                               mock.Mock(return_value="fake")):
            result = manager.reset_task_state(share, state)

            manager._action.assert_called_once_with(
                "reset_task_state", share, {'task_state': state})
            self.assertEqual("fake", result)

    def test_migration_cancel(self):
        share = "fake_share"
        manager = shares.ShareManager(api=fakes.FakeClient())

        with mock.patch.object(manager, "_action",
                               mock.Mock(return_value="fake")):
            result = manager.migration_cancel(share)

            manager._action.assert_called_once_with(
                "migration_cancel", share)
            self.assertEqual("fake", result)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils as osc_utils

from manilaclient import api_versions
from manilaclient.common._i18n import _
from manilaclient.osc import utils


class ListSharePools(command.Lister):
    """List all backend storage pools known to the scheduler (Admin only)."""
    _description = _(
        "List all backend storage pools known to the scheduler (Admin only).")

    def get_parser(self, prog_name):
        parser = super(ListSharePools, self).get_parser(prog_name)
        parser.add_argument(
            "--host",
            metavar="<host>",
            default=None,
            help=_("Filter results by host name. "
                   "Regular expressions are supported.")
        )
        parser.add_argument(
            "--backend",
            metavar="<backend>",
            default=None,
            help=_("Filter results by backend name. "
                   "Regular expressions are supported.")
        )
        parser.add_argument(
            "--pool",
            metavar="<pool>",
            default=None,
            help=_("Filter results by pool name. "
                   "Regular expressions are supported.")
        )
        parser.add_argument(
            "--detail",
            action='store_true',
            default=False,
            help=_("Show detailed information about pools.")
        )
        parser.add_argument(
            "--share-type",
            metavar="<share-type>",
            default=None,
            help=_("Filter results by share type name or ID. "
                   "Available only for microversion >= 2.23")
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        share_type = None
        if parsed_args.share_type:
            if share_client.api_version >= api_versions.APIVersion("2.23"):
                share_type = osc_utils.find_resource(
                    share_client.share_types,
                    parsed_args.share_type).id
            else:
                raise exceptions.CommandError(_(
                    "Filtering results by share type is only available with "
                    "manila API version >= 2.23"))

        search_opts = {
            'host': parsed_args.host,
            'backend': parsed_args.backend,
            'pool': parsed_args.pool,
            'share_type': share_type,
        }

        pools = share_client.pools.list(
            detailed=parsed_args.detail, search_opts=search_opts)

        columns = ["Name", "Host", "Backend", "Pool"]

        if parsed_args.detail:
            columns.append("Capabilities")
            if parsed_args.formatter == 'table':
                for pool in pools:
                    pool._info.update({
                        'capabilities': utils.format_properties(
                            pool.capabilities)})

        data = (osc_utils.get_dict_properties(
            pool._info, columns) for pool in pools)

        return (columns, data)



# Copyright (c) 2015 Clinton Knight.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from manilaclient import base

RESOURCES_PATH = '/scheduler-stats/pools'
RESOURCES_NAME = 'pools'


class Pool(base.Resource):

    def __repr__(self):
        return "<Pool: %s>" % self.name


class PoolManager(base.Manager):
    """Manage :class:`Pool` resources."""
    resource_class = Pool

    def list(self, detailed=True, search_opts=None):
        """Get a list of pools.

        :rtype: list of :class:`Pool`
        """
        query_string = self._build_query_string(search_opts)
        if detailed:
            path = '%(resources_path)s/detail%(query)s' % {
                'resources_path': RESOURCES_PATH,
                'query': query_string
            }
        else:
            path = '%(resources_path)s%(query)s' % {
                'resources_path': RESOURCES_PATH,
                'query': query_string
            }

        return self._list(path, RESOURCES_NAME)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
from unittest import mock

import fixtures
import requests
import testtools


class TestCase(testtools.TestCase):
    TEST_REQUEST_BASE = {
        'verify': True,
    }

    def setUp(self):
        super(TestCase, self).setUp()
        if (os.environ.get('OS_STDOUT_CAPTURE') == 'True' or
                os.environ.get('OS_STDOUT_CAPTURE') == '1'):
            stdout = self.useFixture(fixtures.StringStream('stdout')).stream
            self.useFixture(fixtures.MonkeyPatch('sys.stdout', stdout))
        if (os.environ.get('OS_STDERR_CAPTURE') == 'True' or
                os.environ.get('OS_STDERR_CAPTURE') == '1'):
            stderr = self.useFixture(fixtures.StringStream('stderr')).stream
            self.useFixture(fixtures.MonkeyPatch('sys.stderr', stderr))

    def mock_object(self, obj, attr_name, new_attr=None, **kwargs):
        """Mock an object attribute.

        Use python mock to mock an object attribute
        Mocks the specified objects attribute with the given value.
        Automatically performs 'addCleanup' for the mock.
        """
        if not new_attr:
            new_attr = mock.Mock()
        patcher = mock.patch.object(obj, attr_name, new_attr, **kwargs)
        patcher.start()
        self.addCleanup(patcher.stop)
        return new_attr

    def mock_completion(self):
        patcher = mock.patch(
            'manilaclient.base.Manager.write_to_completion_cache')
        patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch('manilaclient.base.Manager.completion_cache')
        patcher.start()
        self.addCleanup(patcher.stop)


class TestResponse(requests.Response):
    """Class used to wrap requests.Response.

    Class used to wrap requests.Response and provide some
    convenience to initialize with a dict.
    """

    def __init__(self, data):
        self._text = None
        super(TestResponse, self)
        if isinstance(data, dict):
            self.status_code = data.get('status_code', None)
            self.headers = data.get('headers', {})
            self.headers['x-openstack-request-id'] = data.get(
                'x-openstack-request-id', 'fake-request-id')
            # Fake the text attribute to streamline Response creation
            self._text = data.get('text', None)
        else:
            self.status_code = data
            self.headers = {'x-openstack-request-id': 'fake-request-id'}

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    @property
    def text(self):
        return self._text



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from manilaclient.tests.functional.osc import base


class ShareNetworksCLITest(base.OSCClientTestBase):

    def test_openstack_share_network_create(self):
        share_network_name = 'test_create_share_network'
        share_network = self.create_share_network(name=share_network_name)

        self.assertEqual(share_network['name'], share_network_name)

        share_network_list = self.listing_result('share network', 'list')
        self.assertIn(share_network['id'],
                      [item['ID'] for item in share_network_list])

    def test_openstack_share_network_list(self):
        share_network = self.create_share_network()
        share_network_list = self.listing_result('share network', 'list')
        self.assertTableStruct(share_network_list, [
            'ID',
            'Name',
        ])
        self.assertIn(share_network['id'],
                      [item['ID'] for item in share_network_list])

    def test_openstack_share_network_show(self):
        share_network = self.create_share_network()

        result = self.dict_result('share network',
                                  'show %s' % share_network['id'])
        self.assertEqual(share_network['id'], result['id'])

        listing_result = self.listing_result('share network',
                                             'show %s' % share_network['id'])
        self.assertTableStruct(listing_result, [
            'Field',
            'Value'
        ])

    def test_openstack_share_network_delete(self):
        share_network = self.create_share_network(add_cleanup=False)
        share_network_list = self.listing_result('share network', 'list')

        self.assertIn(share_network['id'],
                      [item['ID'] for item in share_network_list])

        self.openstack('share network delete %s' % share_network['id'])
        self.check_object_deleted('share network', share_network['id'])
        share_network_list_after_delete = self.listing_result('share network',
                                                              'list')

        self.assertNotIn(
            share_network['id'],
            [item['ID'] for item in share_network_list_after_delete])

    def test_openstack_share_network_set(self):
        share_network = self.create_share_network()
        self.openstack('share network set %s --name %s'
                       % (share_network['id'], 'new_name'))
        result = self.dict_result('share network', 'show %s' %
                                  share_network['id'])
        self.assertEqual(share_network['id'], result['id'])
        self.assertEqual('new_name', result['name'])

    def test_openstack_share_network_unset(self):
        share_network = self.create_share_network(name='test_name')
        result1 = self.dict_result('share network', 'show %s' %
                                   share_network['id'])
        self.assertEqual(share_network['id'], result1['id'])
        self.assertEqual(share_network['name'], result1['name'])

        self.openstack('share network unset %s --name'
                       % (share_network['id']))
        result2 = self.dict_result('share network', 'show %s' %
                                   share_network['id'])
        self.assertEqual(share_network['id'], result2['id'])
        self.assertEqual('None', result2['name'])



# Copyright 2021 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import logging

from openstackclient.identity import common as identity_common
from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils as oscutils

from manilaclient import api_versions
from manilaclient.common._i18n import _
from manilaclient.common import cliutils

LOG = logging.getLogger(__name__)


class ListShareNetwork(command.Lister):
    _description = _("List share networks")

    def get_parser(self, prog_name):
        parser = super(ListShareNetwork, self).get_parser(prog_name)

        parser.add_argument(
            '--name',
            metavar="<share-network>",
            help=_('Filter share networks by name')
        )
        parser.add_argument(
            '--name~',
            metavar="<share-network-name-pattern>",
            help=_('Filter share networks by name-pattern. Available only '
                   'for microversion >= 2.36.')
        )
        parser.add_argument(
            '--description',
            metavar="<share-network-description>",
            help=_('Filter share networks by description. Available '
                   'only for microversion >= 2.36')
        )
        parser.add_argument(
            '--description~',
            metavar="<share-network-description-pattern>",
            help=_('Filter share networks by description-pattern. Available '
                   'only for microversion >= 2.36.')
        )
        parser.add_argument(
            '--all-projects',
            action='store_true',
            default=False,
            help=_('Include all projects (admin only)'),
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_('Filter share networks by project (name or ID) '
                   '(admin only)')
        )
        identity_common.add_project_domain_option_to_parser(parser)
        parser.add_argument(
            '--created-since',
            metavar='<yyyy-mm-dd>',
            help=_('Filter share networks by date they were created after. '
                   'The date can be in the format yyyy-mm-dd.')
        )
        parser.add_argument(
            '--created-before',
            metavar='<yyyy-mm-dd>',
            help=_('Filter share networks by date they were created before. '
                   'The date can be in the format yyyy-mm-dd.')
        )
        parser.add_argument(
            '--security-service',
            metavar='<security-service>',
            help=_('Filter share networks by the name or ID of a security '
                   'service attached to the network.')
        )
        parser.add_argument(
            '--neutron-net-id',
            metavar='<neutron-net-id>',
            help=_('Filter share networks by the ID of a neutron network.')
        )
        parser.add_argument(
            '--neutron-subnet-id',
            metavar='<neutron-subnet-id>',
            help=_('Filter share networks by the ID of a neutron sub network.')
        )
        parser.add_argument(
            '--network-type',
            metavar='<network-type>',
            help=_('Filter share networks by the type of network. Examples '
                   'include "flat", "vlan", "vxlan", "geneve", etc.')
        )
        parser.add_argument(
            '--segmentation-id',
            metavar='<segmentation-id>',
            help=_('Filter share networks by the segmentation ID of network. '
                   'Relevant only for segmented networks such as "vlan", '
                   '"vxlan", "geneve", etc.')
        )
        parser.add_argument(
            '--cidr',
            metavar='<X.X.X.X/X>',
            help=_('Filter share networks by the CIDR of network.')
        )
        parser.add_argument(
            '--ip-version',
            metavar='4/6',
            choices=['4', '6'],
            help=_('Filter share networks by the IP Version of the network, '
                   'either 4 or 6.')
        )
        parser.add_argument(
            '--detail',
            action='store_true',
            default=False,
            help=_("List share networks with details")
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        identity_client = self.app.client_manager.identity

        columns = ['ID', 'Name']

        if parsed_args.detail:
            columns.extend([
                'Status',
                'Created At',
                'Updated At',
                'Description',
            ])

        project_id = None
        if parsed_args.project:
            project_id = identity_common.find_project(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain).id

        # set value of 'all_tenants' when using project option
        all_tenants = bool(parsed_args.project) or parsed_args.all_projects

        if parsed_args.all_projects:
            columns.append('Project ID')

        search_opts = {
            'all_tenants': all_tenants,
            'project_id': project_id,
            'name': parsed_args.name,
            'created_since': parsed_args.created_since,
            'created_before': parsed_args.created_before,
            'neutron_net_id': parsed_args.neutron_net_id,
            'neutron_subnet_id': parsed_args.neutron_subnet_id,
            'network_type': parsed_args.network_type,
            'segmentation_id': parsed_args.segmentation_id,
            'cidr': parsed_args.cidr,
            'ip_version': parsed_args.ip_version,
            'security_service': parsed_args.security_service,
        }

        if share_client.api_version >= api_versions.APIVersion("2.36"):
            search_opts['name~'] = getattr(parsed_args, 'name~')
            search_opts['description~'] = getattr(parsed_args, 'description~')
            search_opts['description'] = parsed_args.description
        elif (parsed_args.description or getattr(parsed_args, 'name~') or
              getattr(parsed_args, 'description~')):
            raise exceptions.CommandError(
                "Pattern based filtering (name~, description~ and description)"
                " is only available with manila API version >= 2.36")

        data = share_client.share_networks.list(search_opts=search_opts)

        return (
            columns,
            (oscutils.get_item_properties(s, columns) for s in data)
        )


class ShowShareNetwork(command.ShowOne):
    """Display a share network"""
    _description = _("Show details about a share network")

    def get_parser(self, prog_name):
        parser = super(ShowShareNetwork, self).get_parser(prog_name)
        parser.add_argument(
            "share_network",
            metavar="<share-network>",
            help=_("Name or ID of the share network to display")
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        share_network = oscutils.find_resource(
            share_client.share_networks,
            parsed_args.share_network)

        data = share_network._info

        # Add security services information
        security_services = share_client.security_services.list(
            search_opts={'share_network_id': share_network.id}, detailed=False)
        data['security_services'] = [
            {
                'security_service_name': ss.name,
                'security_service_id': ss.id,
            }
            for ss in security_services
        ]

        if parsed_args.formatter == 'table':
            data['share_network_subnets'] = (
                cliutils.convert_dict_list_to_string(
                    data['share_network_subnets'])
            )
            data['security_services'] = cliutils.convert_dict_list_to_string(
                data['security_services']
            )

        data.pop('links', None)

        return self.dict2columns(data)


class CreateShareNetwork(command.ShowOne):
    """Create a share network."""
    _description = _("Create a share network")

    def get_parser(self, prog_name):
        parser = super(CreateShareNetwork, self).get_parser(prog_name)
        parser.add_argument(
            "--name",
            metavar="<share-network>",
            help=_("Add a name to the share network (Optional)")
        )
        parser.add_argument(
            "--description",
            metavar="<description>",
            default=None,
            help=_("Add a description to the share network (Optional).")
        )
        parser.add_argument(
            "--neutron-net-id",
            metavar="<neutron-net-id>",
            default=None,
            help=_("ID of the neutron network that must be associated with "
                   "the share network (Optional). The network specified will "
                   "be associated with the 'default' share network subnet, "
                   "unless 'availability-zone' is also specified.")
        )
        parser.add_argument(
            "--neutron-subnet-id",
            metavar="<neutron-subnet-id>",
            default=None,
            help=_("ID of the neutron sub-network that must be associated "
                   "with the share network (Optional). The subnet specified "
                   "will be associated with the 'default' share network "
                   "subnet, unless 'availability-zone' is also specified.")
        )
        parser.add_argument(
            "--availability-zone",
            metavar="<availability-zone>",
            default=None,
            help=_("Name or ID of the avalilability zone to assign the "
                   "specified network subnet parameters to. Must be used "
                   "in conjunction with 'neutron-net-id' and "
                   "'neutron-subnet-id'. Do not specify this parameter if "
                   "the network must be available across all availability "
                   "zones ('default' share network subnet). Available "
                   "only for microversion >= 2.51.")
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        availability_zone = None
        if (parsed_args.availability_zone and
                share_client.api_version < api_versions.APIVersion("2.51")):
            raise exceptions.CommandError(
                "Availability zone can be specified only with manila API "
                "version >= 2.51")
        elif parsed_args.availability_zone:
            availability_zone = oscutils.find_resource(
                share_client.availability_zones,
                parsed_args.availability_zone).name

        share_network = share_client.share_networks.create(
            name=parsed_args.name,
            description=parsed_args.description,
            neutron_net_id=parsed_args.neutron_net_id,
            neutron_subnet_id=parsed_args.neutron_subnet_id,
            availability_zone=availability_zone,
        )
        share_network_data = share_network._info
        share_network_data.pop('links', None)
        if parsed_args.formatter == 'table':
            share_network_data['share_network_subnets'] = (
                cliutils.convert_dict_list_to_string(
                    share_network_data['share_network_subnets'])
            )
        return self.dict2columns(share_network_data)


class DeleteShareNetwork(command.Command):
    """Delete one or more share networks"""
    _description = _("Delete one or more share networks")

    def get_parser(self, prog_name):
        parser = super(DeleteShareNetwork, self).get_parser(prog_name)
        parser.add_argument(
            "share_network",
            metavar="<share-network>",
            nargs="+",
            help=_("Name or ID of the share network(s) to delete")
        )
        parser.add_argument(
            "--wait",
            action='store_true',
            default=False,
            help=_("Wait for the share network(s) to be deleted")
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        result = 0

        for share_network in parsed_args.share_network:
            try:
                share_network_obj = oscutils.find_resource(
                    share_client.share_networks,
                    share_network)
                share_client.share_networks.delete(
                    share_network_obj)

                if parsed_args.wait:
                    if not oscutils.wait_for_delete(
                            manager=share_client.share_networks,
                            res_id=share_network_obj.id):
                        result += 1
            except Exception as e:
                result += 1
                LOG.error(f"Failed to delete share network with "
                          f"name or ID {share_network}: {e}")

        if result > 0:
            total = len(parsed_args.share_network)
            msg = f"{result} of {total} share networks failed to be deleted."
            raise exceptions.CommandError(msg)


class SetShareNetwork(command.Command):
    """Set share network properties."""
    _description = _("Set share network properties")

    def get_parser(self, prog_name):
        parser = super(SetShareNetwork, self).get_parser(prog_name)
        parser.add_argument(
            "share_network",
            metavar="<share-network>",
            help=_('Name or ID of the share network to set a property for')
        )
        parser.add_argument(
            "--name",
            metavar="<name>",
            default=None,
            help=_("Set a new name to the share network.")
        )
        parser.add_argument(
            "--description",
            metavar="<description>",
            default=None,
            help=_("Set a new description to the share network.")
        )
        parser.add_argument(
            "--status",
            metavar="<status>",
            choices=['active', 'error', 'network_change'],
            help=_("Assign a status to the share network (Admin only). "
                   "Options include : active, error, network_change. "
                   "Available only for microversion >= 2.63.")
        )
        parser.add_argument(
            '--neutron-net-id',
            metavar='<neutron-net-id>',
            help=_("Update the neutron network associated with the default "
                   "share network subnet. If a default share network subnet "
                   "is not present or if the share network is in use, setting "
                   "this will fail.")
        )
        parser.add_argument(
            '--neutron-subnet-id',
            metavar='<neutron-subnet-id>',
            help=_("Update the neutron subnetwork associated with the default "
                   "share network subnet. If a default share network subnet "
                   "is not present or if the share network is in use, setting "
                   "this will fail.")
        )
        parser.add_argument(
            '--current-security-service',
            metavar='<security-service>',
            help=_("Name or ID of a security service that is currently "
                   "associated with a share network that must be replaced. "
                   "Replacing a security service is only available for "
                   "microversions >= 2.63.")
        )
        parser.add_argument(
            '--new-security-service',
            metavar='<security-service>',
            help=_("Name or ID of a security service that must be associated "
                   "with the share network. When replacing a security "
                   "service, the current security service must also be "
                   "provided with the '--current-security-service' option. "
                   "Replacing a security service is only available for "
                   "microversions >= 2.63.")
        )
        parser.add_argument(
            '--check-only',
            action='store_true',
            help=_("Run a dry-run of a security service replacement. "
                   "Available only for microversion >=2.63")
        )
        parser.add_argument(
            '--restart-check',
            action='store_true',
            help=_("Restart a dry-run of a security service "
                   "replacement. Helpful when check results are "
                   "stale. Available only for microversion >=2.63.")
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        result = 0

        share_network = oscutils.find_resource(share_client.share_networks,
                                               parsed_args.share_network)

        kwargs = {}

        # some args are only for newer API micro-version
        if share_client.api_version < api_versions.APIVersion("2.63"):
            new_args = ('status', 'check_only', 'restart_check',
                        'current_security_service')
            invalid_args = [
                arg for arg in new_args
                if getattr(parsed_args, arg) is not None
            ]
            raise exceptions.CommandError(
                f"Use of {' '.join(invalid_args)} is only available for API "
                f"microversions >= 2.63.")

        # If "--check-only" and "--restart-check" are used, a new security
        # service option must be supplied
        if parsed_args.check_only or parsed_args.restart_check:
            if not parsed_args.new_security_service:
                raise exceptions.CommandError(
                    _("Must provide new security service with --check-only "
                      "and --restart-check."))

        if parsed_args.name is not None:
            kwargs['name'] = parsed_args.name
        if parsed_args.description is not None:
            kwargs['description'] = parsed_args.description
        if parsed_args.neutron_net_id is not None:
            kwargs['neutron_net_id'] = parsed_args.neutron_net_id
        if parsed_args.neutron_subnet_id is not None:
            kwargs['neutron_subnet_id'] = parsed_args.neutron_subnet_id

        if kwargs:
            try:
                share_client.share_networks.update(
                    share_network,
                    **kwargs
                )
            except Exception as e:
                result += 1
                LOG.error(f"Failed to set share network properties "
                          f"{kwargs}: {e}")

        if parsed_args.status:
            try:
                share_client.share_networks.reset_state(
                    share_network,
                    parsed_args.status
                )
            except Exception as e:
                result += 1
                LOG.error(f"Failed to update share network status to "
                          f"{parsed_args.status}: {e}")

        new_security_service = current_security_service = None
        if parsed_args.new_security_service:
            new_security_service = oscutils.find_resource(
                share_client.security_services,
                parsed_args.new_security_service)
        if parsed_args.current_security_service:
            current_security_service = oscutils.find_resource(
                share_client.security_services,
                parsed_args.current_security_service)

        if new_security_service and not current_security_service:
            try:
                if parsed_args.check_only:
                    check_result = share_client.share_networks\
                        .add_security_service_check(
                            share_network, new_security_service,
                            reset_operation=parsed_args.restart_check)
                    is_compatible = check_result[1].get('compatible')
                    # NOTE(gouthamr): We're logging to the console here,
                    # because there's no need to print useful
                    # information beyond the result. Use of error
                    # logging is a hack, since other kinds of logs may
                    # not print to console by default.
                    if is_compatible is None:
                        LOG.error("Security service check has been "
                                  "successfully initiated, please retry "
                                  "after some time.")
                    elif is_compatible:
                        LOG.error(
                            f"Security service "
                            f"{parsed_args.new_security_service} can "
                            f"be added to share network "
                            f"{parsed_args.share_network}.")
                    else:
                        LOG.error(
                            f"Security service "
                            f"{parsed_args.new_security_service} cannot "
                            f"be added to share network "
                            f"{parsed_args.share_network}.")
                else:
                    share_client.share_networks.add_security_service(
                        share_network, new_security_service)
            except Exception as e:
                result += 1
                LOG.error(f"Failed to add security service  "
                          f"{parsed_args.new_security_service} to "
                          f"share network {parsed_args.share_network}: {e}")

        if new_security_service and current_security_service:
            try:
                if parsed_args.check_only:
                    check_result = share_client.share_networks.\
                        update_share_network_security_service_check(
                            share_network,
                            current_security_service,
                            new_security_service,
                            reset_operation=parsed_args.restart_check)

                    is_compatible = check_result[1].get('compatible')
                    # NOTE(gouthamr): We're logging to the console here,
                    # because there's no need to print useful
                    # information beyond the result. Use of error
                    # logging is a hack, since other kinds of logs may
                    # not print to console by default.
                    if is_compatible is None:
                        LOG.error("Security service check has been "
                                  "successfully initiated, please retry "
                                  "after some time.")
                    elif is_compatible:
                        LOG.error(
                            f"Security service "
                            f"{parsed_args.current_security_service} can "
                            f"be replaced with security service "
                            f"{parsed_args.new_security_service} on share "
                            f"network {parsed_args.share_network}.")
                    else:
                        LOG.error(
                            f"Security service "
                            f"{parsed_args.current_security_service} cannot "
                            f"be replaced with security service "
                            f"{parsed_args.new_security_service} on share "
                            f"network {parsed_args.share_network}.")
                else:
                    share_client.share_networks.\
                        update_share_network_security_service(
                            share_network,
                            current_security_service,
                            new_security_service)
            except Exception as e:
                result += 1
                LOG.error(f"Failed to update security service "
                          f"{parsed_args.current_security_service} on "
                          f"share network "
                          f"{parsed_args.share_network}: {e}")

        if result > 0:
            raise exceptions.CommandError(_("One or more of the "
                                          "set operations failed"))


class UnsetShareNetwork(command.Command):
    """Unset a share network property."""
    _description = _("Unset a share network property")

    def get_parser(self, prog_name):
        parser = super(UnsetShareNetwork, self).get_parser(prog_name)
        parser.add_argument(
            "share_network",
            metavar="<share-network>",
            help=_("Name or ID of the share network to unset a property of")
        )
        parser.add_argument(
            "--name",
            action='store_true',
            help=_("Unset share network name."),
        )
        parser.add_argument(
            "--description",
            action='store_true',
            help=_("Unset share network description."),
        )
        parser.add_argument(
            "--security-service",
            metavar="<security-service>",
            help=_("Disassociate a security service from the share network. "
                   "This is only possible with unused share networks."),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        share_network = oscutils.find_resource(
            share_client.share_networks,
            parsed_args.share_network)

        security_service = None
        if parsed_args.security_service:
            security_service = oscutils.find_resource(
                share_client.security_services,
                parsed_args.security_service)

        result = 0
        kwargs = {}
        if parsed_args.name:
            # the SDK unsets name if it is an empty string
            kwargs['name'] = ''
        if parsed_args.description:
            # the SDK unsets description if it is an empty string
            kwargs['description'] = ''
        if kwargs:
            try:
                share_client.share_networks.update(
                    share_network,
                    **kwargs
                )
            except Exception as e:
                result += 1
                LOG.error(f"Failed to unset share network properties "
                          f"{kwargs}: {e}")

        if security_service:
            try:
                share_client.share_networks.remove_security_service(
                    share_network, security_service)
            except Exception as e:
                result += 1
                LOG.error(f"Failed to dissociate security service"
                          f"{parsed_args.security_service} from "
                          f"{parsed_args.share_network}: {e}")

        if result > 0:
            raise exceptions.CommandError(_("One or more of the "
                                          "unset operations failed"))



# Copyright 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from manilaclient import api_versions
from manilaclient import base

RESOURCE_PATH_LEGACY = '/os-quota-sets'
RESOURCE_PATH = '/quota-sets'
REPLICA_QUOTAS_MICROVERSION = "2.53"


class QuotaSet(base.Resource):

    @property
    def id(self):
        """Needed by Resource to self-refresh and be indexed."""
        return self.tenant_id

    def update(self, *args, **kwargs):
        self.manager.update(self.tenant_id, *args, **kwargs)


class QuotaSetManager(base.ManagerWithFind):
    resource_class = QuotaSet

    def _check_user_id_and_share_type_args(self, user_id, share_type):
        if user_id and share_type:
            raise ValueError(
                "'user_id' and 'share_type' values are mutually exclusive. "
                "one or both should be unset.")

    def _do_get(self, tenant_id, user_id=None, share_type=None, detail=False,
                resource_path=RESOURCE_PATH):
        self._check_user_id_and_share_type_args(user_id, share_type)
        if hasattr(tenant_id, 'tenant_id'):
            tenant_id = tenant_id.tenant_id

        if detail:
            query = '/detail'
        else:
            query = ''

        if user_id and share_type:
            query = '%s?user_id=%s&share_type=%s' % (
                query, user_id, share_type)
        elif user_id:
            query = '%s?user_id=%s' % (query, user_id)
        elif share_type:
            query = '%s?share_type=%s' % (query, share_type)
        data = {
            "resource_path": resource_path,
            "tenant_id": tenant_id,
        }

        url = ("%(resource_path)s/%(tenant_id)s" + query) % data
        return self._get(url, "quota_set")

    @api_versions.wraps("1.0", "2.6")
    def get(self, tenant_id, user_id=None, detail=False):
        return self._do_get(tenant_id, user_id,
                            resource_path=RESOURCE_PATH_LEGACY)

    @api_versions.wraps("2.7", "2.24")  # noqa
    def get(self, tenant_id, user_id=None, detail=False):  # noqa
        return self._do_get(tenant_id, user_id,
                            resource_path=RESOURCE_PATH)

    @api_versions.wraps("2.25", "2.38")  # noqa
    def get(self, tenant_id, user_id=None, detail=False):  # noqa
        return self._do_get(tenant_id, user_id, detail=detail,
                            resource_path=RESOURCE_PATH)

    @api_versions.wraps("2.39")  # noqa
    def get(self, tenant_id, user_id=None, share_type=None, detail=False):   # noqa
        return self._do_get(
            tenant_id, user_id, share_type=share_type, detail=detail,
            resource_path=RESOURCE_PATH)

    def _do_update(self, tenant_id, shares=None, snapshots=None,
                   gigabytes=None, snapshot_gigabytes=None,
                   share_networks=None,
                   force=None, user_id=None, share_type=None,
                   share_groups=None, share_group_snapshots=None,
                   share_replicas=None, replica_gigabytes=None,
                   per_share_gigabytes=None,
                   resource_path=RESOURCE_PATH):
        self._check_user_id_and_share_type_args(user_id, share_type)
        body = {
            'quota_set': {
                'tenant_id': tenant_id,
                'shares': shares,
                'snapshots': snapshots,
                'gigabytes': gigabytes,
                'snapshot_gigabytes': snapshot_gigabytes,
                'share_networks': share_networks,
                'share_groups': share_groups,
                'share_group_snapshots': share_group_snapshots,
                'force': force,
                'share_replicas': share_replicas,
                'replica_gigabytes': replica_gigabytes,
                'per_share_gigabytes': per_share_gigabytes,
            },
        }

        for key in list(body['quota_set']):
            if body['quota_set'][key] is None:
                body['quota_set'].pop(key)
        data = {
            "resource_path": resource_path,
            "tenant_id": tenant_id,
            "user_id": user_id,
            "st": share_type,
        }
        if user_id:
            url = '%(resource_path)s/%(tenant_id)s?user_id=%(user_id)s' % data
        elif share_type:
            url = '%(resource_path)s/%(tenant_id)s?share_type=%(st)s' % data
        else:
            url = "%(resource_path)s/%(tenant_id)s" % data

        return self._update(url, body, 'quota_set')

    @api_versions.wraps("1.0", "2.6")
    def update(self, tenant_id, shares=None, snapshots=None, gigabytes=None,
               snapshot_gigabytes=None, share_networks=None, force=None,
               user_id=None):
        return self._do_update(
            tenant_id, shares, snapshots, gigabytes, snapshot_gigabytes,
            share_networks, force, user_id, resource_path=RESOURCE_PATH_LEGACY,
        )

    @api_versions.wraps("2.7", "2.38")  # noqa
    def update(self, tenant_id, shares=None, snapshots=None, gigabytes=None,  # noqa
               snapshot_gigabytes=None, share_networks=None, force=None,
               user_id=None):
        return self._do_update(
            tenant_id, shares, snapshots, gigabytes, snapshot_gigabytes,
            share_networks, force, user_id, resource_path=RESOURCE_PATH,
        )

    def _validate_st_and_sn_in_same_request(self, share_type, share_networks):
        if share_type and share_networks:
            raise ValueError(
                "'share_networks' quota can be set only for project or user, "
                "not share type.")

    @api_versions.wraps("2.39", "2.39")  # noqa
    def update(self, tenant_id, user_id=None, share_type=None,  # noqa
               shares=None, snapshots=None, gigabytes=None,
               snapshot_gigabytes=None, share_networks=None, force=None):
        self._validate_st_and_sn_in_same_request(share_type, share_networks)
        return self._do_update(
            tenant_id, shares, snapshots, gigabytes, snapshot_gigabytes,
            share_networks, force, user_id,
            share_type=share_type,
            resource_path=RESOURCE_PATH,
        )

    @api_versions.wraps("2.40", "2.52")  # noqa
    def update(self, tenant_id, user_id=None, share_type=None,  # noqa
               shares=None, snapshots=None, gigabytes=None,
               snapshot_gigabytes=None, share_networks=None,
               share_groups=None, share_group_snapshots=None,
               force=None):
        self._validate_st_and_sn_in_same_request(share_type, share_networks)
        return self._do_update(
            tenant_id, shares, snapshots, gigabytes, snapshot_gigabytes,
            share_networks, force, user_id,
            share_type=share_type,
            share_groups=share_groups,
            share_group_snapshots=share_group_snapshots,
            resource_path=RESOURCE_PATH,
        )

    @api_versions.wraps(REPLICA_QUOTAS_MICROVERSION, "2.61")  # noqa
    def update(self, tenant_id, user_id=None, share_type=None,  # noqa
               shares=None, snapshots=None, gigabytes=None,
               snapshot_gigabytes=None, share_networks=None,
               share_groups=None, share_group_snapshots=None,
               share_replicas=None, replica_gigabytes=None, force=None):
        self._validate_st_and_sn_in_same_request(share_type, share_networks)
        return self._do_update(
            tenant_id, shares, snapshots, gigabytes, snapshot_gigabytes,
            share_networks, force, user_id,
            share_type=share_type,
            share_groups=share_groups,
            share_group_snapshots=share_group_snapshots,
            share_replicas=share_replicas,
            replica_gigabytes=replica_gigabytes,
            resource_path=RESOURCE_PATH
        )

    @api_versions.wraps("2.62")  # noqa
    def update(self, tenant_id, user_id=None, share_type=None,  # noqa
               shares=None, snapshots=None, gigabytes=None,
               snapshot_gigabytes=None, share_networks=None,
               share_groups=None, share_group_snapshots=None,
               share_replicas=None, replica_gigabytes=None, force=None,
               per_share_gigabytes=None):
        self._validate_st_and_sn_in_same_request(share_type, share_networks)
        return self._do_update(
            tenant_id, shares, snapshots, gigabytes, snapshot_gigabytes,
            share_networks, force, user_id,
            share_type=share_type,
            share_groups=share_groups,
            share_group_snapshots=share_group_snapshots,
            share_replicas=share_replicas,
            replica_gigabytes=replica_gigabytes,
            per_share_gigabytes=per_share_gigabytes,
            resource_path=RESOURCE_PATH
        )

    @api_versions.wraps("1.0", "2.6")
    def defaults(self, tenant_id):
        return self._get(
            "%(resource_path)s/%(tenant_id)s/defaults" % {
                "resource_path": RESOURCE_PATH_LEGACY, "tenant_id": tenant_id},
            "quota_set")

    @api_versions.wraps("2.7")  # noqa
    def defaults(self, tenant_id):  # noqa
        return self._get(
            "%(resource_path)s/%(tenant_id)s/defaults" % {
                "resource_path": RESOURCE_PATH, "tenant_id": tenant_id},
            "quota_set")

    def _do_delete(self, tenant_id, user_id=None, share_type=None,
                   resource_path=RESOURCE_PATH):
        self._check_user_id_and_share_type_args(user_id, share_type)
        data = {
            "resource_path": resource_path,
            "tenant_id": tenant_id,
            "user_id": user_id,
            "st": share_type,
        }
        if user_id:
            url = '%(resource_path)s/%(tenant_id)s?user_id=%(user_id)s' % data
        elif share_type:
            url = '%(resource_path)s/%(tenant_id)s?share_type=%(st)s' % data
        else:
            url = '%(resource_path)s/%(tenant_id)s' % data
        self._delete(url)

    @api_versions.wraps("1.0", "2.6")
    def delete(self, tenant_id, user_id=None):
        return self._do_delete(
            tenant_id, user_id, resource_path=RESOURCE_PATH_LEGACY)

    @api_versions.wraps("2.7", "2.38")  # noqa
    def delete(self, tenant_id, user_id=None):  # noqa
        return self._do_delete(tenant_id, user_id, resource_path=RESOURCE_PATH)

    @api_versions.wraps("2.39")  # noqa
    def delete(self, tenant_id, user_id=None, share_type=None):   # noqa
        return self._do_delete(
            tenant_id, user_id, share_type, resource_path=RESOURCE_PATH)
